"""Abstract baseclass for all distributions."""
import logging
import numpy

import chaospy

from .utils import check_dependencies, report_on_exception


class Distribution(object):
    """Baseclass for all probability distributions."""

    __array_priority__ = 9000
    """Numpy override variable."""

    interpret_as_integer = False
    """
    Flag indicating that return value from the methods sample, and inv
    should be interpreted as integers instead of floating point.
    """

    @property
    def stochastic_dependent(self):
        """True if distribution contains stochastically dependent components."""
        return any(len(deps) > 1 for deps in self._dependencies)

    def __init__(
            self,
            parameters,
            dependencies,
            rotation=None,
            exclusion=None,
            repr_args=None,
    ):
        """
        Distribution initializer.

        In addition to assigning some object variables, also checks for
        some consistency issues.

        Args:
            parameters (Optional[Distribution[str, Union[ndarray, Distribution]]]):
                Collection of model parameters.
            dependencies (Optional[Sequence[Set[int]]]):
                Dependency identifiers. One collection for each dimension.
            rotation (Optional[Sequence[int]]):
                The order of which to resolve dependencies.
            exclusion (Optional[Sequence[int]]):
                Distributions that has been "taken out of play" and
                therefore can not be reused other places in the
                dependency hierarchy.
            repr_args (Optional[Sequence[str]]):
                Positional arguments to place in the object string
                representation. The repr output will then be:
                `<class name>(<arg1>, <arg2>, ...)`.

        Raises:
            StochasticallyDependentError:
                For dependency structures that can not later be
                rectified. This include under-defined
                distributions, and inclusion of distributions that
                should be exclusion.

        """
        assert isinstance(parameters, dict)
        self._parameters = parameters
        self._dependencies = list(dependencies)
        if rotation is None:
            rotation = sorted(enumerate(self._dependencies), key=lambda x: len(x[1]))
            rotation = [key for key, _ in rotation]
        rotation = list(rotation)
        assert len(set(rotation)) == len(dependencies)
        assert min(rotation) == 0
        assert max(rotation) == len(dependencies)-1
        self._rotation = rotation
        if exclusion is None:
            exclusion = set()
        self._exclusion = set(exclusion)
        if repr_args is None:
            repr_args = ("{}={}".format(key, self._parameters[key])
                         for key in sorted(self._parameters))
        self._repr_args = list(repr_args)
        self._mom_cache = {(0,)*len(dependencies): 1.}
        self._ttr_cache = {}
        self._indices = {}

        self._all_dependencies = {dep for deps in self._dependencies for dep in deps}
        if len(self._all_dependencies) < len(dependencies):
            raise chaospy.StochasticallyDependentError(
                "%s is an under-defined probability distribution." % self)

        for key, param in list(parameters.items()):
            if isinstance(param, Distribution):
                if self._all_dependencies.intersection(param._exclusion):
                    raise chaospy.StochasticallyDependentError((
                        "%s contains dependencies that can not also exist "
                        "other places in the dependency hierarchy") % param)
                self._exclusion.update(param._exclusion)
            else:
                self._parameters[key] = numpy.asarray(param)

    def get_parameters(self, idx, cache, assert_numerical=True):
        """Get distribution parameters."""
        del assert_numerical
        out = self._parameters.copy()
        assert isinstance(cache, dict)
        if idx is not None:
            assert not isinstance(idx, dict), idx
            assert idx == int(idx), idx
            if "idx" in out:
                assert out["idx"] == idx, (idx, out)
        assert "cache" not in out
        out["cache"] = cache
        out["idx"] = idx
        return out

    @property
    def lower(self):
        """Lower bound for the distribution."""
        cache = {}
        out = numpy.zeros(len(self))
        for idx in self._rotation:
            out[idx] = self._get_lower(idx, cache=cache)
        return out

    @report_on_exception
    def _get_lower(self, idx, cache):
        """In-processes function for getting lower bounds."""
        if (idx, self) in cache:
            return cache[idx, self][0]
        parameters = self.get_parameters(idx, cache, assert_numerical=False)
        out = self._lower(**parameters)
        assert not isinstance(out, Distribution), (self, out)
        out = numpy.atleast_1d(out)
        assert out.ndim == 1, (self, out, cache)
        cache[idx, self] = (out, None)
        return out

    def _lower(self, **kwargs):
        """Backend lower bound."""
        raise chaospy.UnsupportedFeature("lower not supported")

    @property
    def upper(self):
        """Upper bound for the distribution."""
        cache = {}
        out = numpy.zeros(len(self))
        for idx in self._rotation:
            out[idx] = self._get_upper(idx, cache=cache)
        return out

    @report_on_exception
    def _get_upper(self, idx, cache):
        """In-processes function for getting upper bounds."""
        if (idx, self) in cache:
            return cache[idx, self][0]
        parameters = self.get_parameters(idx, cache, assert_numerical=False)
        out = self._upper(**parameters)
        assert not isinstance(out, Distribution), (self, out)
        out = numpy.atleast_1d(out)
        assert out.ndim == 1, (self, out, cache)
        cache[idx, self] = (out, None)
        size = max([elem[0].size for elem in cache.values()])
        assert all([elem[0].size in (1, size) for elem in cache.values()])
        return out

    def _upper(self, **kwargs):
        """Backend upper bound."""
        raise chaospy.UnsupportedFeature("lower not supported")

    def fwd(self, x_data):
        """
        Forward Rosenblatt transformation.

        Args:
            x_data (numpy.ndarray):
                Location for the distribution function. ``x_data.shape`` must
                be compatible with distribution shape.

        Returns:
            (numpy.ndarray):
                Evaluated distribution function values, where
                ``out.shape==x_data.shape``.
        """
        logger = logging.getLogger(__name__)
        check_dependencies(self)
        x_data = numpy.asfarray(x_data)
        shape = x_data.shape
        x_data = x_data.reshape(len(self), -1)
        cache = {}
        q_data = numpy.zeros(x_data.shape)
        for idx in self._rotation:
            q_data[idx] = self._get_fwd(x_data[idx], idx, cache)

        indices = (q_data > 1) | (q_data < 0)
        if numpy.any(indices):
            logger.debug("%s.fwd: %d/%d outputs out of bounds",
                         self, numpy.sum(indices), len(indices))
            q_data = numpy.clip(q_data, a_min=0, a_max=1)

        q_data = q_data.reshape(shape)
        return q_data

    @report_on_exception
    def _get_fwd(self, x_data, idx, cache):
        """In-process function for getting cdf-values."""
        logger = logging.getLogger(__name__)
        if (idx, self) in cache:
            return cache[idx, self][1]
        lower = numpy.broadcast_to(self._get_lower(idx, cache=cache.copy()), x_data.shape)
        upper = numpy.broadcast_to(self._get_upper(idx, cache=cache.copy()), x_data.shape)
        parameters = self.get_parameters(idx, cache, assert_numerical=True)
        ret_val = self._cdf(x_data, **parameters)
        assert not isinstance(ret_val, Distribution), (self, ret_val)
        out = numpy.zeros(x_data.shape)
        out[:] = ret_val
        indices = x_data < lower
        if numpy.any(indices):
            logger.debug("%s.fwd: %d/%d inputs below bounds",
                         self, numpy.sum(indices), len(indices))
            out = numpy.where(indices, 0, out)
        indices = x_data > upper
        if numpy.any(indices):
            logger.debug("%s.fwd: %d/%d inputs above bounds",
                         self, numpy.sum(indices), len(indices))
            out = numpy.where(indices, 1, out)
        assert numpy.all((out >= 0) | (out <= 1))

        cache[idx, self] = (x_data, out)
        assert out.ndim == 1, (self, out, cache)
        return out

    def cdf(self, x_data):
        """
        Cumulative distribution function.

        Note that chaospy only supports cumulative distribution functions for
        stochastically independent distributions.

        Args:
            x_data (numpy.ndarray):
                Location for the distribution function. Assumes that
                ``len(x_data) == len(distribution)``.

        Returns:
            (numpy.ndarray):
                Evaluated distribution function values, where output has shape
                ``x_data.shape`` in one dimension and ``x_data.shape[1:]`` in
                higher dimensions.
        """
        check_dependencies(self)
        if self.stochastic_dependent:
            raise chaospy.StochasticallyDependentError(
                "Cumulative distribution does not support dependencies.")
        x_data = numpy.asarray(x_data)
        if self.interpret_as_integer:
            x_data = x_data+0.5
        q_data = self.fwd(x_data)
        if len(self) > 1:
            q_data = numpy.prod(q_data, 0)
        return q_data

    def inv(self, q_data, max_iterations=100, tollerance=1e-5):
        """
        Inverse Rosenblatt transformation.

        If possible the transformation is done analytically. If not possible,
        transformation is approximated using an algorithm that alternates
        between Newton-Raphson and binary search.

        Args:
            q_data (numpy.ndarray):
                Probabilities to be inverse. If any values are outside ``[0,
                1]``, error will be raised. ``q_data.shape`` must be compatible
                with distribution shape.
            max_iterations (int):
                If approximation is used, this sets the maximum number of
                allowed iterations in the Newton-Raphson algorithm.
            tollerance (float):
                If approximation is used, this set the error tolerance level
                required to define a sample as converged.

        Returns:
            (numpy.ndarray):
                Inverted probability values where
                ``out.shape == q_data.shape``.
        """
        logger = logging.getLogger(__name__)
        check_dependencies(self)
        q_data = numpy.asfarray(q_data)
        assert numpy.all((q_data >= 0) & (q_data <= 1)), "sanitize your inputs!"
        shape = q_data.shape
        q_data = q_data.reshape(len(self), -1)
        cache = {}
        x_data = numpy.zeros(q_data.shape)
        for idx in self._rotation:
            x_data[idx] = self._get_inv(q_data[idx], idx, cache)

        x_data = x_data.reshape(shape)
        return x_data

    @report_on_exception
    def _get_inv(self, q_data, idx, cache):
        """In-process function for getting ppf-values."""
        logger = logging.getLogger(__name__)
        assert numpy.all(q_data <= 1) and numpy.all(q_data >= 0)
        assert q_data.ndim == 1
        if (idx, self) in cache:
            return cache[idx, self][0]
        lower = numpy.broadcast_to(self._get_lower(idx, cache=cache.copy()), q_data.shape)
        upper = numpy.broadcast_to(self._get_upper(idx, cache=cache.copy()), q_data.shape)
        try:
            parameters = self.get_parameters(idx, cache, assert_numerical=True)
            ret_val = self._ppf(q_data, **parameters)
        except chaospy.UnsupportedFeature:
            ret_val = chaospy.approximate_inverse(
                self, idx, q_data, cache=cache)
        assert not isinstance(ret_val, Distribution), (self, ret_val)

        out = numpy.zeros(q_data.shape)
        out[:] = ret_val

        indices = out < lower
        if numpy.any(indices):
            logger.debug("%s.inv: %d/%d outputs below bounds",
                         self, numpy.sum(indices), len(indices))
            out = numpy.where(indices, lower, out)

        indices = out > upper
        if numpy.any(indices):
            logger.debug("%s.inv: %d/%d outputs above bounds",
                         self, numpy.sum(indices), len(indices))
            out = numpy.where(indices, upper, out)

        assert out.ndim == 1
        cache[idx, self] = (out, q_data)
        assert out.ndim == 1, (self, out, cache)
        return out

    def _ppf(self, xloc, **kwargs):
        raise chaospy.UnsupportedFeature(
            "%s: does not support analytical ppf." % self)

    def pdf(self, x_data, decompose=False):
        """
        Probability density function.

        If possible the density will be calculated analytically. If not
        possible, it will be approximated by approximating the one-dimensional
        derivative of the forward Rosenblatt transformation and multiplying the
        component parts. Note that even if the distribution is multivariate,
        each component of the Rosenblatt is one-dimensional.

        Args:
            x_data (numpy.ndarray):
                Location for the density function. If multivariate,
                `len(x_data) == len(self)` is required.
            decompose (bool):
                Decompose multivariate probability density `p(x), p(y|x), ...`
                instead of multiplying them together into `p(x, y, ...)`.

        Returns:
            (numpy.ndarray):
                Evaluated density function evaluated in `x_data`. If decompose,
                `output.shape == x_data.shape`, else if multivariate the first
                dimension is multiplied together.

        Example:
            >>> chaospy.Gamma(2).pdf([1, 2, 3, 4, 5]).round(3)
            array([0.368, 0.271, 0.149, 0.073, 0.034])
            >>> dist = chaospy.Iid(chaospy.Normal(0, 1), 2)
            >>> grid = numpy.mgrid[-1.5:2, -1.5:2]
            >>> dist.pdf(grid).round(3)
            array([[0.017, 0.046, 0.046, 0.017],
                   [0.046, 0.124, 0.124, 0.046],
                   [0.046, 0.124, 0.124, 0.046],
                   [0.017, 0.046, 0.046, 0.017]])
            >>> dist.pdf(grid, decompose=True).round(3)
            array([[[0.13 , 0.13 , 0.13 , 0.13 ],
                    [0.352, 0.352, 0.352, 0.352],
                    [0.352, 0.352, 0.352, 0.352],
                    [0.13 , 0.13 , 0.13 , 0.13 ]],
            <BLANKLINE>
                   [[0.13 , 0.352, 0.352, 0.13 ],
                    [0.13 , 0.352, 0.352, 0.13 ],
                    [0.13 , 0.352, 0.352, 0.13 ],
                    [0.13 , 0.352, 0.352, 0.13 ]]])

        """
        logger = logging.getLogger(__name__)
        check_dependencies(self)
        x_data = numpy.asfarray(x_data)
        shape = x_data.shape
        x_data = x_data.reshape(len(self), -1)
        f_data = numpy.zeros(x_data.shape)
        cache = {}
        for idx in self._rotation:
            f_data[idx] = self._get_pdf(x_data[idx], idx, cache)

        f_data = f_data.reshape(shape)
        if len(self) > 1 and not decompose:
            f_data = numpy.prod(f_data, 0)
        return f_data

    @report_on_exception
    def _get_pdf(self, x_data, idx, cache):
        """In-process function for getting pdf-values."""
        logger = logging.getLogger(__name__)
        assert x_data.ndim == 1
        if (idx, self) in cache:
            return cache[idx, self][1]
        lower = numpy.broadcast_to(self._get_lower(idx, cache=cache.copy()), x_data.shape)
        upper = numpy.broadcast_to(self._get_upper(idx, cache=cache.copy()), x_data.shape)
        parameters = self.get_parameters(idx, cache, assert_numerical=True)
        try:
            ret_val = self._pdf(x_data, **parameters)
        except chaospy.UnsupportedFeature:
            ret_val = chaospy.approximate_density(self, idx, x_data, cache)
        assert not isinstance(ret_val, Distribution), (self, ret_val)

        out = numpy.zeros(x_data.shape)
        out[:] = ret_val

        indices = (x_data < lower) | (x_data > upper)
        if numpy.any(indices):
            logger.debug("%s.fwd: %d/%d inputs out of bounds",
                         self, numpy.sum(indices), len(indices))
            logger.debug("%s[%s]: %s - %s - %s", self, idx, lower, x_data, upper)
            out = numpy.where(indices, 0, ret_val)

        if self in cache:
            out = numpy.where(x_data == cache[self][0], out, 0)
        cache[idx, self] = (x_data, out)
        assert out.ndim == 1, (self, out, cache)
        return out

    def _pdf(self, xloc, **kwargs):
        raise chaospy.UnsupportedFeature(
            "%s: does not support analytical pdf." % self)

    def sample(self, size=(), rule="random", antithetic=None):
        """
        Create pseudo-random generated samples.

        By default, the samples are created using standard (pseudo-)random
        samples. However, if needed, the samples can also be created by either
        low-discrepancy sequences, and/or variance reduction techniques.

        Changing the sampling scheme, use the following ``rule`` flag:

        +----------------------+-----------------------------------------------+
        | key                  | Description                                   |
        +======================+===============================================+
        | ``chebyshev``        | Roots of first order Chebyshev polynomials.   |
        +----------------------+-----------------------------------------------+
        | ``nested_chebyshev`` | Chebyshev nodes adjusted to ensure nested.    |
        +----------------------+-----------------------------------------------+
        | ``korobov``          | Korobov lattice.                              |
        +----------------------+-----------------------------------------------+
        | ``random``           | Classical (Pseudo-)Random samples.            |
        +----------------------+-----------------------------------------------+
        | ``grid``             | Regular spaced grid.                          |
        +----------------------+-----------------------------------------------+
        | ``nested_grid``      | Nested regular spaced grid.                   |
        +----------------------+-----------------------------------------------+
        | ``latin_hypercube``  | Latin hypercube samples.                      |
        +----------------------+-----------------------------------------------+
        | ``sobol``            | Sobol low-discrepancy sequence.               |
        +----------------------+-----------------------------------------------+
        | ``halton``           | Halton low-discrepancy sequence.              |
        +----------------------+-----------------------------------------------+
        | ``hammersley``       | Hammersley low-discrepancy sequence.          |
        +----------------------+-----------------------------------------------+

        All samples are created on the ``[0, 1]``-hypercube, which then is
        mapped into the domain of the distribution using the inverse Rosenblatt
        transformation.

        Args:
            size (numpy.ndarray):
                The size of the samples to generate.
            rule (str):
                Indicator defining the sampling scheme.
            antithetic (bool, numpy.ndarray):
                If provided, will be used to setup antithetic variables. If
                array, defines the axes to mirror.

        Returns:
            (numpy.ndarray):
                Random samples with shape ``(len(self),)+self.shape``.
        """
        check_dependencies(self)
        size_ = numpy.prod(size, dtype=int)
        dim = len(self)
        if dim > 1:
            if isinstance(size, (tuple, list, numpy.ndarray)):
                shape = (dim,) + tuple(size)
            else:
                shape = (dim, size)
        else:
            shape = size

        from chaospy.distributions import sampler
        out = sampler.generator.generate_samples(
            order=size_, domain=self, rule=rule, antithetic=antithetic)
        try:
            out = out.reshape(shape)
        except:
            if len(self) == 1:
                out = out.flatten()
            else:
                out = out.reshape(dim, int(out.size/dim))

        if self.interpret_as_integer:
            out = numpy.round(out).astype(int)
        return out

    def mom(self, K, allow_approx=True, **kws):
        """
        Raw statistical moments.

        Creates non-centralized raw moments from the random variable. If
        analytical options can not be utilized, Monte Carlo integration
        will be used.

        Args:
            K (numpy.ndarray):
                Index of the raw moments. k.shape must be compatible with
                distribution shape.  Sampling scheme when performing Monte
                Carlo
            rule (str):
                rule for estimating the moment if the analytical method fails.
            antithetic (numpy.ndarray):
                List of bool. Represents the axes to mirror using antithetic
                variable during MCI.

        Returns:
            (numpy.ndarray):
                Shapes are related through the identity
                ``k.shape == dist.shape+k.shape``.
        """
        logger = logging.getLogger(__name__)
        K = numpy.asarray(K, dtype=int)
        assert numpy.all(K >= 0)
        shape = K.shape
        dim = len(self)

        if dim > 1:
            shape = shape[1:]

        size = int(K.size/dim)
        K = K.reshape(dim, size)
        try:
            out = [self._get_mom(kdata) for kdata in K.T]
            logger.debug("%s: PDF calculated successfully", str(self))
        except chaospy.UnsupportedFeature:
            if allow_approx:
                logger.info(
                    "%s: has stochastic dependencies; "
                    "Approximating moments with quadrature.", str(self))
                out = [chaospy.approximate_moment(self, kdata) for kdata in K.T]
            else:
                raise
        out = numpy.array(out)
        assert out.size == numpy.prod(shape), (out, shape)
        return out.reshape(shape)

    @report_on_exception
    def _get_mom(self, kdata):
        """In-process function for getting moments."""
        if tuple(kdata) in self._mom_cache:
            return self._mom_cache[tuple(kdata)]
        parameters = self.get_parameters(idx=None, cache={}, assert_numerical=False)
        assert "idx" not in parameters, (self, parameters)
        ret_val = float(self._mom(kdata, **parameters))
        assert not isinstance(ret_val, Distribution), (self, ret_val)
        self._mom_cache[tuple(kdata)] = ret_val
        return ret_val

    def _mom(self, kloc, **kwargs):
        raise chaospy.UnsupportedFeature(
            "moments not supported for this distribution")

    def ttr(self, kloc):
        """
        Three terms relation's coefficient generator.

        Args:
            k (numpy.ndarray, int):
                The order of the coefficients.

        Returns:
            (Recurrence coefficients):
                Where out[0] is the first (A) and out[1] is the second
                coefficient With ``out.shape==(2,)+k.shape``.
        """
        check_dependencies(self)
        kloc = numpy.asarray(kloc, dtype=int)
        shape = kloc.shape
        kloc = kloc.reshape(len(self), -1)
        out = numpy.zeros((2,)+kloc.shape)
        for idy, kloc_ in enumerate(kloc.T):
            for idx in range(len(self)):
                out[:, idx, idy] = self._get_ttr(kloc_[idx], idx)
        return out.reshape((2,)+shape)

    @report_on_exception
    def _get_ttr(self, kdata, idx):
        """In-process function for getting TTR-values."""
        if (idx, kdata) in self._ttr_cache:
            return self._ttr_cache[idx, kdata]
        parameters = self.get_parameters(idx, cache={}, assert_numerical=True)
        alpha, beta = self._ttr(kdata, **parameters)
        assert not isinstance(alpha, Distribution), (self, alpha)
        assert not isinstance(beta, Distribution), (self, beta)
        alpha = numpy.asfarray(alpha).item()
        beta = numpy.asfarray(beta).item()
        self._ttr_cache[idx, kdata] = (alpha, beta)
        return alpha, beta

    @report_on_exception
    def _get_cache(self, idx, cache, get=None):
        if (idx, self) in cache:
            out = cache[idx, self]
        else:
            out = self._cache(idx, cache)
        if not isinstance(out, Distribution) and get is not None:
            assert get in (0, 1)
            out = out[get]
        return out

    @report_on_exception
    def _get_cache_1(self, idx, cache):
        """
        In-process function for getting cached values.

        Each time a distribution has been processed, the input and output
        values are stored in the cache.
        This checks if a distribution has been processed before and return a
        cache value if it is.

        The cached values are as follows:

        -----------  -------------
        Context      Content
        -----------  -------------
        pdf          Input values
        cdf/fwd      Input values
        ppf/inv      Output values
        lower/upper  Output values
        -----------  -------------

        Args:
            cache (Dict[Distribution, Tuple[numpy.ndarray, numpy.ndarray]]):
                Collection of cached values. Keys are distributions that has
                been processed earlier, values consist of up to two cache
                value.

        Returns:
            (numpy.ndarray, Distribution):
                The content of the first cache, if any. Else return self.
        """
        if self in cache:
            out = cache[idx, self]
        else:
            out = self._cache(idx, cache)
        if not isinstance(out, Distribution):
            out = out[0]
        return out

    @report_on_exception
    def _get_cache_2(self, idx, cache):
        """
        In-process function for getting cached values.

        Each time a distribution has been processed, the input and output
        values are stored in the cache.
        This checks if a distribution has been processed before and return a
        cache value if it is.

        The cached values are as follows:

        -----------  -------------
        Context      Content
        -----------  -------------
        pdf          Output values
        cdf/fwd      Output values
        ppf/inv      Input values
        lower/upper  N/A
        -----------  -------------

        Args:
            cache (Dict[Distribution, Tuple[numpy.ndarray, numpy.ndarray]]):
                Collection of cached values. Keys are distributions that has
                been processed earlier, values consist of up to two cache
                value.

        Returns:
            (numpy.ndarray, Distribution):
                The content of the second cache, if any. Else return self.
        """
        if self in cache:
            out = cache[idx, self]
        else:
            out = self._cache(idx, cache)
        if not isinstance(out, Distribution):
            out = out[1]
        return out

    def _cache(self, idx, cache):
        """Backend function of retrieving cache values."""
        return self

    def __getitem__(self, index):
        if isinstance(index, numpy.number):
            assert index.dtype == int
            index = int(index)
        if isinstance(index, int):
            if not -len(self) < index < len(self):
                raise IndexError("index out of bounds: %s" % index)
            if index < 0:
                index += len(self)
            return chaospy.ItemDistribution(int(index), self)
        if isinstance(index, slice):
            start = 0 if index.start is None else index.start
            stop = len(self) if index.stop is None else index.stop
            step = 1 if index.step is None else index.step
            return chaospy.J(*[self[idx] for idx in range(start, stop, step)])
        raise IndexError("unrecognized key: %s" % repr(index))

    def __iter__(self):
        if isinstance(self, (chaospy.ItemDistribution, chaospy.SimpleDistribution)):
            raise TypeError("index of an index not possible.")
        for idx in range(len(self)):
            yield self[idx]

    def __len__(self):
        """Distribution length."""
        return len(self._dependencies)

    def __repr__(self):
        """Distribution repr function."""
        args = ", ".join([str(arg) for arg in self._repr_args])
        return "{}({})".format(self.__class__.__name__, args)

    def __str__(self):
        """Distribution str function."""
        return repr(self)

    def __add__(self, X):
        """Y.__add__(X) <==> X+Y"""
        return chaospy.Add(self, X)

    def __radd__(self, X):
        """Y.__radd__(X) <==> Y+X"""
        return chaospy.Add(self, X)

    def __sub__(self, X):
        """Y.__sub__(X) <==> X-Y"""
        return chaospy.Add(self, -X)

    def __rsub__(self, X):
        """Y.__rsub__(X) <==> Y-X"""
        return chaospy.Add(X, -self)

    def __neg__(self):
        """X.__neg__() <==> -X"""
        return chaospy.Negative(self)

    def __mul__(self, X):
        """Y.__mul__(X) <==> X*Y"""
        return chaospy.Multiply(self, X)

    def __rmul__(self, X):
        """Y.__rmul__(X) <==> Y*X"""
        return chaospy.Multiply(X, self)

    def __div__(self, X):
        """Y.__div__(X) <==> Y/X"""
        return chaospy.Multiply(self, X**-1)

    def __rdiv__(self, X):
        """Y.__rdiv__(X) <==> X/Y"""
        return chaospy.Multiply(X, self**-1)

    def __floordiv__(self, X):
        """Y.__floordiv__(X) <==> Y/X"""
        return chaospy.Multiply(self, X**-1)

    def __rfloordiv__(self, X):
        """Y.__rfloordiv__(X) <==> X/Y"""
        return chaospy.Multiply(X, self**-1)

    def __truediv__(self, X):
        """Y.__truediv__(X) <==> Y/X"""
        return chaospy.Multiply(self, X**-1)

    def __rtruediv__(self, X):
        """Y.__rtruediv__(X) <==> X/Y"""
        return chaospy.Multiply(X, self**-1)

    def __pow__(self, X):
        """Y.__pow__(X) <==> Y**X"""
        return chaospy.Power(self, X)

    def __rpow__(self, X):
        """Y.__rpow__(X) <==> X**Y"""
        return chaospy.Power(X, self)

    def __eq__(self, other):
        if not isinstance(other, Distribution):
            return False
        if len(other) != len(self):
            return False
        if len(self) > 1:
            return all([self == other for self, other in zip(self, other)])
        if isinstance(self, chaospy.ItemDistribution) and isinstance(other, chaospy.ItemDistribution):
            return (self._parameters["index"] == other._parameters["index"] and
                    self._parameters["parent"] is other._parameters["parent"])

        return self is other

    def __hash__(self):
        return id(self)



"""Collection of mathematical operators."""
from .addition import Add
from .multiply import Multiply
from .negative import Negative
from .power import Power
from .truncation import Trunc
from .logarithm import Log, Log10, Logn

from .joint import J
from .iid import Iid



r"""
Clayton Copula.

The calculation of the derivative of the ``iphi`` function:

.. math::
  \begin{eqnarray}
    iphi(u)     &= (1+\theta*u)^(-1/\theta) \\
    iphi'(u)    &= \theta*(-1/\theta)*(1+\theta*u)^(-1/\theta-1)
                = \theta*\sigma(1+\theta*u, 1, \theta) \\
    iphi''(u)   &= \theta*(-1/\theta)*theta*(-1/\theta-1)*(1+\theta*u)^{-1/\theta-2}
                = \theta^2*\sigma(1+\theta*u, 2, \theta) \\
                & \dots \\
    iphi^(n)(u) &= \theta^n*\prod_{d=0}^{n-1}(-1/\theta-d)*(1+\theta*u)^{-1/\theta-n}
                = \theta^n*sigma(1+theta*u, n)
  \end{eqnarray}
"""
import numpy
from scipy import special

from ..baseclass import CopulaDistribution
from .archimedean import Archimedean


class clayton(Archimedean):
    """Clayton copula."""

    def _phi(self, t_loc, theta):
        return (t_loc**-theta-1)/theta

    def _delta_phi(self, t_loc, theta):
        return -t_loc**(-theta-1)

    def _inverse_phi(self, u_loc, theta, order):
        return theta**order*self._sigma(1+theta*u_loc, theta, order)


class Clayton(CopulaDistribution):
    """
    Clayton Copula.

    Args:
        dist (Distribution):
            The distribution to wrap
        theta (float):
            Copula parameter. Required to be above 0.

    Examples:
        >>> distribution = chaospy.Clayton(
        ...     chaospy.Iid(chaospy.Uniform(-1, 1), 2), theta=2)
        >>> distribution
        Clayton(Iid(Uniform(lower=-1, upper=1), 2), theta=2)
        >>> samples = distribution.sample(3)
        >>> samples.round(4)
        array([[ 0.3072, -0.77  ,  0.9006],
               [ 0.2736, -0.3015,  0.1539]])
        >>> distribution.pdf(samples).round(4)
        array([0.3679, 0.1855, 0.2665])
        >>> distribution.fwd(samples).round(4)
        array([[0.6536, 0.115 , 0.9503],
               [0.4822, 0.8725, 0.2123]])
        >>> mesh = numpy.meshgrid([.4, .5, .6], [.4, .5, .6])
        >>> distribution.inv(mesh).round(4)
        array([[[-0.2   ,  0.    ,  0.2   ],
                [-0.2   ,  0.    ,  0.2   ],
                [-0.2   ,  0.    ,  0.2   ]],
        <BLANKLINE>
               [[-0.2008, -0.0431,  0.0945],
                [-0.0746,  0.0928,  0.2329],
                [ 0.0636,  0.2349,  0.3713]]])

    """

    def __init__(self, dist, theta=2.):
        assert theta > 0
        return super(Clayton, self).__init__(
            dist=dist,
            trans=clayton(len(dist), theta=theta),
            repr_args=[dist, "theta=%s" % theta],
        )



"""
Collection of modified Cholesky decomposion algorithm.

Original Python translation by:
M. M. Forbes <michael.forbes@gmail.com>
"""
from .bastos_ohagen_ import bastos_ohagen
from .gill_king_ import gill_king
from .gill_murray_wright_ import gill_murray_wright
from .schnabel_eskow_ import schnabel_eskow



"""Truncated exponential distribution."""
import numpy
from scipy import special

from ..baseclass import SimpleDistribution, ShiftScaleDistribution


class truncexpon(SimpleDistribution):
    """Truncated exponential distribution."""

    def __init__(self, b):
        super(truncexpon, self).__init__(dict(b=b))

    def _pdf(self, x, b):
        return numpy.exp(-x)/(1-numpy.exp(-b))

    def _cdf(self, x, b):
        return (1.0-numpy.exp(-x))/(1-numpy.exp(-b))

    def _ppf(self, q, b):
        return -numpy.log(1-q+q*numpy.exp(-b))

    def _lower(self, b):
        return 0.

    def _upper(self, b):
        return b


class TruncExponential(ShiftScaleDistribution):
    """
    Truncated exponential distribution.

    Args:
        upper (float, Distribution):
            Location of upper threshold
        scale (float, Distribution):
            Scaling parameter in the exponential distribution
        shift (float, Distribution):
            Location parameter

    Examples:
        >>> distribution = chaospy.TruncExponential(1.5)
        >>> distribution
        TruncExponential(1.5)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([0.   , 0.169, 0.372, 0.628, 0.972, 1.5  ])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([1.287, 1.087, 0.887, 0.687, 0.487, 0.287])
        >>> distribution.sample(4).round(3)
        array([0.709, 0.094, 1.34 , 0.469])

    """

    def __init__(self, upper=1, scale=1, shift=0):
        super(TruncExponential, self).__init__(
            dist=truncexpon((upper-shift)*1./scale),
            scale=scale,
            shift=shift,
            repr_args=[upper],
        )



"""Binomial probability distribution."""
from functools import wraps
import numpy
from scipy import special

from ..baseclass import SimpleDistribution
from ..operators import J


class binomial(SimpleDistribution):
    """
    Binomial probability distribution.

    Point density:
        comb(N, x) p^x (1-p)^{N-x}      x in {0, 1, ..., N}

    Examples:
        >>> distribution = chaospy.Binomial(3, 0.5)
        >>> distribution
        Binomial(3, 0.5)
        >>> xloc = numpy.arange(4)
        >>> distribution.pdf(xloc).round(4)
        array([0.125, 0.375, 0.375, 0.125])
        >>> distribution.cdf(xloc).round(4)
        array([0.125, 0.5  , 0.875, 1.   ])
        >>> distribution.fwd([-0.5, -0.49, 0, 0.49, 0.5]).round(4)
        array([0.    , 0.0013, 0.0625, 0.1238, 0.125 ])
        >>> uloc = numpy.linspace(0, 1, 8)
        >>> uloc.round(2)
        array([0.  , 0.14, 0.29, 0.43, 0.57, 0.71, 0.86, 1.  ])
        >>> distribution.inv(uloc).round(2)
        array([-0.5 ,  0.55,  0.93,  1.31,  1.69,  2.07,  2.45,  3.5 ])
        >>> distribution.sample(10)
        array([2, 1, 0, 2, 2, 2, 2, 3, 3, 0])
        >>> distribution.mom([1, 2, 3]).round(4)
        array([1.5 , 3.  , 6.75])
        >>> distribution.ttr([0, 1, 2, 3]).round(4)
        array([[1.5 , 1.5 , 1.5 , 1.5 ],
               [1.  , 0.75, 1.  , 0.75]])

    """
    interpret_as_integer = True

    def __init__(self, size, prob):
        super(binomial, self).__init__(
            parameters=dict(size=size, prob=prob),
            repr_args=[size, prob],
        )

    def _cdf(self, x_data, size, prob):
        size = numpy.round(size)
        x_data = x_data-0.5

        floor = numpy.zeros(x_data.shape)
        indices = x_data >= 0
        floor[indices] = special.bdtr(numpy.floor(x_data[indices]), size, prob)

        ceil = numpy.ones(x_data.shape)
        indices = x_data <= size
        ceil[indices] = special.bdtr(numpy.ceil(x_data[indices]), size, prob)
        ceil[numpy.isnan(ceil)] = 0  # left edge case

        offset = x_data-numpy.floor(x_data)
        out = floor*(1-offset) + ceil*offset
        return out

    def _pdf(self, x_data, size, prob):
        x_data = numpy.round(x_data)
        return special.comb(size, x_data)*prob**x_data*(1-prob)**(size-x_data)

    def _lower(self, size, prob):
        return -0.5

    def _upper(self, size, prob):
        return numpy.round(size)+0.5

    def _mom(self, k_data, size, prob):
        x_data = numpy.arange(int(size)+1, dtype=int)
        return numpy.sum(x_data**k_data*self._pdf(
            x_data, size=numpy.floor(size), prob=prob))

    def _ttr(self, k_data, size, prob):
        """Krawtchouk rule."""
        from chaospy.quadrature import discretized_stieltjes
        abscissas = numpy.arange(0, numpy.floor(size)+1)
        weights = self._pdf(abscissas, size, prob)
        (alpha, beta), _, _ = discretized_stieltjes(k_data, [abscissas], weights)
        return alpha[0, -1], beta[0, -1]


class Binomial(J):

    def __init__(self, size, prob):
        dist = binomial(size, prob)
        super(Binomial, self).__init__(dist)
        self._repr_args = [size, prob]



"""Generalized gamma distribution."""
import numpy
from scipy import special

from ..baseclass import SimpleDistribution, ShiftScaleDistribution


class generalized_gamma(SimpleDistribution):
    """Generalized gamma distribution."""

    def __init__(self, a, c):
        super(generalized_gamma, self).__init__(dict(a=a, c=c))

    def _pdf(self, x, a, c):
        return abs(c)*numpy.exp(
            (c*a-1)*numpy.log(numpy.clip(x, 1e-15, None))-x**c-special.gammaln(a))

    def _cdf(self, x, a, c):
        val = special.gammainc(a, x**c)
        cond = c+0*val
        return numpy.where(cond > 0, val, 1-val)

    def _ppf(self, q, a, c):
        val = numpy.where(c > 0, q, 1-q)
        return special.gammaincinv(a, val)**(1./c)

    def _lower(self, a, c):
        return 0.

    def _upper(self, a, c):
        cond = c > 0
        val = numpy.where(cond, 1-1e-15, 1e-15)
        return special.gammaincinv(a, val)**(1./c)

    def _mom(self, k, a, c):
        return special.gamma((c+k)*1./a)/special.gamma(c*1./a)


class GeneralizedGamma(ShiftScaleDistribution):
    """
    Generalized gamma distribution

    Args:
        shape1 (float, Distribution):
            Shape parameter 1
        shape2 (float, Distribution):
            Shape parameter 2
        scale (float, Distribution):
            Scaling parameter
        shift (float, Distribution):
            Location parameter

    Examples:
        >>> distribution = chaospy.GeneralizedGamma(1.5, 15)
        >>> distribution
        GeneralizedGamma(1.5, 15)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([0.   , 0.955, 0.995, 1.026, 1.058, 1.271])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0.   , 3.82 , 6.033, 6.76 , 5.555, 0.   ])
        >>> distribution.sample(4).round(3)
        array([1.034, 0.928, 1.095, 1.009])
        >>> distribution.mom(1).round(3)
        4.591

    """

    def __init__(self, shape1, shape2, scale=1, shift=0):
        super(GeneralizedGamma, self).__init__(
            dist=generalized_gamma(shape1, shape2),
            scale=scale,
            shift=shift,
            repr_args=[shape1, shape2],
        )



"""Exponential power distribution."""
import numpy

from ..baseclass import SimpleDistribution, ShiftScaleDistribution


class exponential_power(SimpleDistribution):
    """Exponential power distribution."""

    def __init__(self, b=1):
        super(exponential_power, self).__init__(dict(b=b))

    def _pdf(self, x, b):
        xbm1 = x**(b-1.0)
        xb = xbm1*x
        return numpy.exp(1)*b*xbm1*numpy.exp(xb-numpy.exp(xb))

    def _cdf(self, x, b):
        return -numpy.expm1(-numpy.expm1(x**b))

    def _ppf(self, q, b):
        return pow(numpy.log1p(-numpy.log1p(-q)), 1./b)

    def _lower(self, b):
        del b
        return 0.

    def _upper(self, b):
        return 3.6**(1./b)


class ExponentialPower(ShiftScaleDistribution):
    """
    Exponential power distribution.

    Also known as Generalized error distribution and Generalized normal
    distribution version 1.

    Args:
        shape (float, Distribution):
            Shape parameter
        scale (float, Distribution):
            Scaling parameter
        shift (float, Distribution):
            Location parameter

    Examples:
        >>> distribution = chaospy.ExponentialPower(1.5)
        >>> distribution
        ExponentialPower(1.5)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([0.   , 0.344, 0.554, 0.751, 0.973, 2.349])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0.   , 0.86 , 1.012, 0.996, 0.772, 0.   ])
        >>> distribution.sample(4).round(3)
        array([0.805, 0.237, 1.243, 0.635])

    """

    def __init__(self, shape=1, scale=1, shift=0):
        super(ExponentialPower, self).__init__(
            dist=exponential_power(shape),
            scale=scale,
            shift=shift,
            repr_args=[shape],
        )



from .utils import *
from .distribution import Distribution
from .slice_ import ItemDistribution
from .simple import SimpleDistribution
from .copula import CopulaDistribution
from .mean_covariance import MeanCovarianceDistribution
from .shift_scale import ShiftScaleDistribution
from .lower_upper import LowerUpperDistribution
from .operator import OperatorDistribution



"""Create Lagrange polynomials."""
import numpy
from scipy.special import comb
import numpoly


def lagrange_polynomial(abscissas, graded=True, reverse=True, sort=None):
    """
    Create Lagrange polynomials.

    Args:
        abscissas (numpy.ndarray):
            Sample points where the Lagrange polynomials shall be defined.
        graded (bool):
            Graded sorting, meaning the indices are always sorted by the index
            sum. E.g. ``q0**2*q1**2*q2**2`` has an exponent sum of 6, and will
            therefore be consider larger than both ``q0**2*q1*q2``,
            ``q0*q1**2*q2`` and ``q0*q1*q2**2``, which all have exponent sum of
            5.
        reverse (bool):
            Reverse lexicographical sorting meaning that ``q0*q1**3`` is
            considered bigger than ``q0**3*q1``, instead of the opposite.

    Example:
        >>> chaospy.lagrange_polynomial([-10, 10]).round(4)
        polynomial([-0.05*q0+0.5, 0.05*q0+0.5])
        >>> chaospy.lagrange_polynomial([-1, 0, 1]).round(4)
        polynomial([0.5*q0**2-0.5*q0, -q0**2+1.0, 0.5*q0**2+0.5*q0])
        >>> poly = chaospy.lagrange_polynomial([[1, 0, 1], [0, 1, 2]])
        >>> poly.round(4)
        polynomial([-0.5*q1+0.5*q0+0.5, -q0+1.0, 0.5*q1+0.5*q0-0.5])
        >>> poly([1, 0, 1], [0, 1, 2]).round(4)
        array([[1., 0., 0.],
               [0., 1., 0.],
               [0., 0., 1.]])
        >>> nodes = numpy.array([[ 0.17,  0.15,  0.17,  0.19],
        ...                      [14.94, 16.69, 16.69, 16.69]])
        >>> poly = chaospy.lagrange_polynomial(nodes)  # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
            ...
        LinAlgError: Lagrange abscissas resulted in invertible matrix
    """
    abscissas = numpy.asfarray(abscissas)
    if len(abscissas.shape) == 1:
        abscissas = abscissas.reshape(1, abscissas.size)
    dim, size = abscissas.shape

    order = 1
    while comb(order+dim, dim) < size:
        order += 1

    indices = numpoly.glexindex(0, order+1, dimensions=dim,
                                graded=graded, reverse=reverse)[:size]
    idx, idy = numpy.mgrid[:size, :size]

    matrix = numpy.prod(abscissas.T[idx]**indices[idy], -1)
    det = numpy.linalg.det(matrix)
    if det == 0:
        raise numpy.linalg.LinAlgError(
            "Lagrange abscissas resulted in invertible matrix")

    names = numpoly.variable(dim).names
    vec = numpoly.monomial(
        0, order+1, names=names, graded=graded, reverse=reverse)[:size]

    coeffs = numpy.zeros((size, size))

    if size == 1:
        out = numpoly.monomial(
            0, 1, names=names, graded=graded, reverse=reverse)*abscissas.item()

    elif size == 2:
        coeffs = numpy.linalg.inv(matrix)
        out = numpoly.sum(vec*(coeffs.T), 1)

    else:
        for i in range(size):
            if i%2 != 0:
                k = 1
            else:
                k=0
            for j in range(size):
                if k%2 == 0:
                    coeffs[i, j] += numpy.linalg.det(matrix[1:, 1:])
                else:
                    if size%2 == 0:
                        coeffs[i, j] += -numpy.linalg.det(matrix[1:, 1:])
                    else:
                        coeffs[i, j] += numpy.linalg.det(matrix[1:, 1:])
                matrix = numpy.roll(matrix, -1, axis=0) 
                k += 1
            matrix = numpy.roll(matrix, -1, axis=1) 
        coeffs /= det
        out = numpoly.sum(vec*(coeffs.T), 1)

    return out



"""Fisk or Log-logistic distribution."""
from ..baseclass import SimpleDistribution, ShiftScaleDistribution


class fisk(SimpleDistribution):
    """Fisk or Log-logistic distribution."""

    def __init__(self, c=1):
        super(fisk, self).__init__(dict(c=c))

    def _pdf(self, x, c):
        return c*x**(c-1.)/(1+x**c)**2

    def _cdf(self, x, c):
        return 1./(1+x**-c)

    def _ppf(self, q, c):
        return (q**(-1.0)-1)**(-1.0/c)

    def _lower(self, c):
        return 0.

    def _upper(self, c):
        return 1e12**(1./c)


class Fisk(ShiftScaleDistribution):
    """
    Fisk or Log-logistic distribution.

    Args:
        shape (float, Distribution):
            Shape parameter
        scale (float, Distribution):
            Scaling parameter
        shift (float, Distribution):
            Location parameter

    Examples:
        >>> distribution = chaospy.Fisk(5)
        >>> distribution
        Fisk(5)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([  0.   ,   0.758,   0.922,   1.084,   1.32 , 251.189])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0.   , 1.056, 1.301, 1.107, 0.606, 0.   ])
        >>> distribution.sample(4).round(3)
        array([1.135, 0.665, 1.804, 0.986])

    """
    def __init__(self, shape=1, scale=1, shift=0):
        super(Fisk, self).__init__(
            dist=fisk(shape),
            scale=scale,
            shift=shift,
            repr_args=[shape],
        )



import numpy
from scipy.special import comb
import numpoly
import chaospy

from .distribution import Distribution


class ShiftScaleDistribution(Distribution):
    """
    Shift-Scaling transformation.

    Linear transforms any distribution of the form `A*X+b` where A is a scaling
    matrix and `b` is a shift vector.

    Args:
        dist (Distribution):
            The underlying distribution to be scaled.
        shift (float, Sequence[float], Distribution):
            Mean vector.
        scale (float, Sequence[float], Distribution):
            Covariance matrix or variance vector if scale is a 1-d vector.
            If omitted, assumed to be 1.
        rotation (Sequence[int], Sequence[Sequence[bool]]):
            The order of which to resolve conditionals. Either as a sequence of
            column rotations, or as a permutation matrix.
            Defaults to `range(len(distribution))` which is the same as
            `p(x0), p(x1|x0), p(x2|x0,x1), ...`.

    """

    def __init__(
            self,
            dist,
            shift=0,
            scale=1,
            rotation=None,
            repr_args=None,
    ):
        assert isinstance(dist, Distribution), "'dist' should be a distribution"
        if repr_args is None:
            repr_args = dist._repr_args[:]
        repr_args += chaospy.format_repr_kwargs(scale=(scale, 1))
        repr_args += chaospy.format_repr_kwargs(shift=(shift, 0))
        length = len(dist) if len(dist) > 1 else None
        dependencies, parameters, rotation = chaospy.declare_dependencies(
            distribution=self,
            parameters=dict(shift=shift, scale=scale),
            rotation=rotation,
            length=length,
        )
        super(ShiftScaleDistribution, self).__init__(
            parameters=parameters,
            rotation=rotation,
            dependencies=dependencies,
            repr_args=repr_args,
        )
        self._dist = dist
        permute = numpy.zeros((len(self._rotation), len(self._rotation)), dtype=int)
        permute[numpy.arange(len(self._rotation), dtype=int), self._rotation] = 1
        self._permute = permute

    def get_parameters(self, idx, cache, assert_numerical=True):

        shift = self._parameters["shift"]
        if isinstance(shift, Distribution):
            shift = shift._get_cache(idx, cache, get=0)
        elif idx is not None and len(shift) > 1:
            shift = shift[idx]

        scale = self._parameters["scale"]
        if isinstance(scale, Distribution):
            scale = scale._get_cache(idx, cache, get=0)
        elif idx is not None and len(scale) > 1:
            scale = scale[idx]

        assert not assert_numerical or not (isinstance(shift, Distribution) or
                                            isinstance(scale, Distribution))

        return dict(dist=self._dist, shift=shift, scale=scale,
                    parameters=self._dist.get_parameters(
                        idx, cache, assert_numerical=assert_numerical))

    def _ppf(self, qloc, dist, shift, scale, parameters):
        return dist._ppf(qloc, **parameters)*scale+shift

    def _cdf(self, xloc, dist, shift, scale, parameters):
        return dist._cdf((xloc-shift)/scale, **parameters)

    def _pdf(self, xloc, dist, shift, scale, parameters):
        return dist._pdf((xloc-shift)/scale, **parameters)/scale

    def _mom(self, kloc, dist, shift, scale, parameters):
        del parameters
        poly = numpoly.variable(len(self))
        poly = numpoly.sum(scale*poly, axis=-1)+shift
        poly = numpoly.set_dimensions(numpoly.prod(poly**kloc), len(self))
        out = sum(dist._get_mom(key)*coeff
                  for key, coeff in zip(poly.exponents, poly.coefficients))
        return out

    def _ttr(self, kloc, dist, shift, scale, parameters):
        coeff0, coeff1 = dist._ttr(kloc, **parameters)
        coeff0 = coeff0*scale+shift
        coeff1 = coeff1*scale*scale
        return coeff0, coeff1

    def _lower(self, dist, shift, scale, parameters):
        return dist._lower(**parameters)*scale+shift

    def _upper(self, dist, shift, scale, parameters):
        return dist._upper(**parameters)*scale+shift



"""
Create Latin Hyper-cube samples.

Example usage
-------------

Normal usage::
    >>> chaospy.create_latin_hypercube_samples(order=4, dim=2).round(4)
    array([[0.6634, 0.2788, 0.9876, 0.1205],
           [0.4681, 0.0531, 0.5102, 0.8493]])
"""
import numpy


def create_latin_hypercube_samples(order, dim=1):
    """
    Latin Hypercube sampling.

    Args:
        order (int):
            The order of the latin hyper-cube. Defines the number of samples.
        dim (int):
            The number of dimensions in the latin hyper-cube.

    Returns (numpy.ndarray):
        Latin hyper-cube with ``shape == (dim, order)``.
    """
    randoms = numpy.random.random(order*dim).reshape((dim, order))
    for dim_ in range(dim):
        perm = numpy.random.permutation(order)  # pylint: disable=no-member
        randoms[dim_] = (perm + randoms[dim_])/order
    return randoms



"""Alpha probability distribution."""
import numpy
from scipy import special

import chaospy
from ..baseclass import SimpleDistribution, ShiftScaleDistribution


class alpha(SimpleDistribution):
    """Standard Alpha distribution."""

    def __init__(self, a=1):
        super(alpha, self).__init__(dict(a=a))

    def _cdf(self, x, a):
        return special.ndtr(a-1./x)/special.ndtr(a)

    def _ppf(self, q, a):
        out = 1.0/(a-special.ndtri(q*special.ndtr(a)))
        return numpy.where(q == 1, self._upper(a), out)

    def _pdf(self, x, a):
        return numpy.where(
            x == 0, 0, numpy.e**(-.5*(a-1./x)**2)/
                (numpy.sqrt(2*numpy.pi)*x**2*special.ndtr(a))
        )

    def _lower(self, a):
        return 0.

    def _upper(self, a):
        return 1./(a-special.ndtri((1-1e-10)*special.ndtr(a)))


class Alpha(ShiftScaleDistribution):
    """
    Alpha distribution.

    Args:
        shape (float, Distribution):
            Shape parameter
        scale (float, Distribution):
            Scale Parameter
        shift (float, Distribution):
            Location of lower threshold

    Examples:
        >>> distribution = chaospy.Alpha(6)
        >>> distribution
        Alpha(6)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([ 0.   ,  0.146,  0.16 ,  0.174,  0.194, 63.709])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([ 0.   , 13.104, 15.108, 12.759,  7.449,  0.   ])
        >>> distribution.sample(4).round(3)
        array([0.178, 0.139, 0.23 , 0.165])

    """

    def __init__(self, shape=1, scale=1, shift=0):
        super(Alpha, self).__init__(
            dist=alpha(shape),
            scale=scale,
            shift=shift,
            repr_args=[shape],
        )



"""Auto correlation function."""
import numpy

from .pearson import Corr


def Acf(poly, dist, N=None, **kws):
    """
    Auto-correlation function.

    Args:
        poly (numpoly.ndpoly):
            Polynomial of interest. Must have ``len(poly) > N``.
        dist (Distribution):
            Defines the space the correlation is taken on.
        N (int):
            The number of time steps apart included. If omitted set to
            ``len(poly)/2+1``.

    Returns:
        (numpy.ndarray) :
            Auto-correlation of ``poly`` with shape ``(N,)``. Note that by
            definition ``Q[0]=1``.

    Examples:
        >>> poly = chaospy.prange(10)[1:]
        >>> dist = chaospy.Uniform()
        >>> chaospy.Acf(poly, dist, 5).round(4)
        array([1.    , 0.9915, 0.9722, 0.9457, 0.9127])

    """
    if N is None:
        N = len(poly)/2 + 1

    corr = Corr(poly, dist, **kws)
    out = numpy.empty(N)

    for n in range(N):
        out[n] = numpy.mean(corr.diagonal(n), 0)

    return out



"""
There are three ways to create a multivariate probability distribution in
``chaospy``: Using the joint constructor
:class:`~chaospy.distributions.operators.joint.J`, the identical independent
distribution constructor: :class:`~chaospy.distributions.operators.iid.Iid`,
and to one of the pre-constructed multivariate distribution defined in
:ref:`listdistributions`.

Constructing a multivariate probability distribution can be done using the
:func:`~chaospy.distributions.operators.joint.J` constructor. E.g.::

    >>> distribution = chaospy.J(chaospy.Normal(), chaospy.Uniform())
    >>> distribution
    J(Normal(mu=0, sigma=1), Uniform())

The created multivariate distribution behaves much like the univariate case::

    >>> mesh = numpy.mgrid[0.25:0.75:3j, 0.25:0.75:3j]
    >>> mesh
    array([[[0.25, 0.25, 0.25],
            [0.5 , 0.5 , 0.5 ],
            [0.75, 0.75, 0.75]],
    <BLANKLINE>
           [[0.25, 0.5 , 0.75],
            [0.25, 0.5 , 0.75],
            [0.25, 0.5 , 0.75]]])
    >>> distribution.cdf(mesh).round(4)
    array([[0.1497, 0.2994, 0.449 ],
           [0.1729, 0.3457, 0.5186],
           [0.1933, 0.3867, 0.58  ]])
    >>> distribution.pdf(mesh).round(4)
    array([[0.3867, 0.3867, 0.3867],
           [0.3521, 0.3521, 0.3521],
           [0.3011, 0.3011, 0.3011]])
    >>> distribution.sample(6, rule="halton").round(4)
    array([[-1.1503,  0.3186, -0.3186,  1.1503, -1.5341,  0.1573],
           [ 0.4444,  0.7778,  0.2222,  0.5556,  0.8889,  0.037 ]])
    >>> distribution.mom([[2, 4, 6], [1, 2, 3]]).round(10)
    array([0.5 , 1.  , 3.75])

"""
import logging
import numpy
import chaospy

from ..baseclass import Distribution


class J(Distribution):
    """
    Joint random variable.

    Args:
        args (chaospy.Distribution):
            Distribution to join together.
        rotation (Optional[Sequence[int]]):
            The order of how the joint should be evaluated.
    """

    def __init__(self, *args, **kwargs):
        dists = []
        repr_args = []
        owners = {}
        index = []
        dependencies = []
        idx = 0
        for arg in args:
            assert isinstance(arg, Distribution)
            if isinstance(arg, J):
                repr_args.extend(arg._repr_args)
                owners.update({idx+idx_: value
                               for idx_, value in arg._owners.items()})
                index += list(range(idx, idx+len(args)))
                idx += len(arg)
            elif len(arg) > 1:
                repr_args.append(arg)
                owners.update({idx2: (idx1, arg)
                               for idx1, idx2 in enumerate(range(idx, idx+len(arg)))})
                index += [idx]*len(arg)
                idx += 1
            else:
                repr_args.append(arg)
                owners[idx] = (0, arg)
                index += [idx]
                idx += 1
            dependencies += [dep.copy() for dep in arg._dependencies]

        self.interpret_as_integer = all([dist.interpret_as_integer
                                         for (_, dist) in owners.values()])
        rotation = kwargs.pop("rotation", None)
        assert not kwargs, "'rotation' is the only allowed keyword."
        parameters = {"_%03d" % idx: dist for idx, dist in enumerate(repr_args)}
        parameters["index"] = numpy.asarray(index)

        super(J, self).__init__(
            parameters=parameters,
            dependencies=dependencies,
            rotation=rotation,
            repr_args=repr_args,
        )
        self._owners = owners

    def get_parameters(self, idx, cache, assert_numerical=True):
        del assert_numerical  # joint is never numerical on its own.
        parameters = super(J, self).get_parameters(
            idx, cache, assert_numerical=False)
        if idx is None:
            return dict(index=parameters["index"])
        idx, dist = self._owners[idx]
        return dict(idx=idx, dist=dist, cache=cache)

    def _lower(self, idx, dist, cache):
        """
        Example:
            >>> dist = chaospy.J(chaospy.Uniform(), chaospy.Normal())
            >>> dist.lower.round(4)
            array([ 0.  , -8.22])
            >>> d0 = chaospy.Uniform()
            >>> dist = chaospy.J(d0, d0+chaospy.Uniform())
            >>> dist.lower
            array([0., 0.])
        """
        return dist._get_lower(idx, cache)

    def _upper(self, idx, dist, cache):
        """
        Example:
            >>> dist = chaospy.J(chaospy.Uniform(), chaospy.Normal())
            >>> dist.upper.round(4)
            array([1.  , 8.22])
            >>> d0 = chaospy.Uniform()
            >>> dist = chaospy.J(d0, d0+chaospy.Uniform())
            >>> dist.upper
            array([1., 2.])
        """
        return dist._get_upper(idx, cache)

    def _cdf(self, xloc, idx, dist, cache):
        """
        Examples:
            >>> dist = chaospy.J(chaospy.Uniform(), chaospy.Normal())
            >>> dist.fwd([[0, 0.5, 1], [-1, 0, 1]]).round(4)
            array([[0.    , 0.5   , 1.    ],
                   [0.1587, 0.5   , 0.8413]])
            >>> d0 = chaospy.Uniform()
            >>> dist = chaospy.J(d0, d0+chaospy.Uniform())
            >>> dist.fwd([[0, 0.5, 1], [0.5, 1, 1.5]]).round(4)
            array([[0. , 0.5, 1. ],
                   [0.5, 0.5, 0.5]])

        """
        return dist._get_fwd(xloc, idx, cache)

    def _pdf(self, xloc, idx, dist, cache):
        """
        Example:
            >>> dist = chaospy.J(chaospy.Uniform(), chaospy.Normal())
            >>> dist.pdf([[-0.5, 0.5, 1.5], [-1, 0, 1]]).round(4)
            array([0.    , 0.3989, 0.    ])
            >>> d0 = chaospy.Uniform()
            >>> dist = chaospy.J(d0, d0+chaospy.Uniform())
            >>> dist.pdf([[-0.5, 0.5, 1.5], [0, 1, 2]]).round(4)
            array([0., 1., 0.])
        """
        return dist._get_pdf(xloc, idx, cache)

    def _ppf(self, qloc, idx, dist, cache):
        """
        Example:
            >>> dist1 = chaospy.J(chaospy.Uniform(), chaospy.Normal())
            >>> xloc = dist1.inv([[0.1, 0.2, 0.3], [0.3, 0.3, 0.4]])
            >>> xloc.round(4)
            array([[ 0.1   ,  0.2   ,  0.3   ],
                   [-0.5244, -0.5244, -0.2533]])
            >>> dist2 = chaospy.Uniform()
            >>> joint = chaospy.J(dist2, dist2+chaospy.Uniform())
            >>> joint.inv([[0.1, 0.2, 0.3], [0.3, 0.3, 0.4]]).round(4)
            array([[0.1, 0.2, 0.3],
                   [0.4, 0.5, 0.7]])

        """
        return dist._get_inv(qloc, idx, cache)

    def _mom(self, kloc, index):
        """
        Example:
            >>> d0 = chaospy.Uniform()
            >>> dist = chaospy.J(d0, d0+chaospy.Uniform())
            >>> dist.mom([1, 1]).round(4)
            0.5833
            >>> dist = chaospy.J(chaospy.Uniform(), chaospy.Normal())
            >>> dist.mom([[0, 0, 1], [0, 1, 1]]).round(4)
            array([1., 0., 0.])
        """
        output = 1.
        for idx1 in range(len(self._owners)):
            _, dist1 = self._owners[idx1]
            for idx2 in range(idx1+1, len(self._owners)):
                _, dist2 = self._owners[idx2]
                if chaospy.shares_dependencies(dist1, dist2):
                    raise chaospy.UnsupportedFeature(
                        "Shared dependencies across joint")

        kloc = kloc[self._rotation]
        for unique_idx in numpy.unique(index[self._rotation]):
            output *= self._owners[unique_idx][1]._get_mom(kloc[index == unique_idx])
        return output

    def _ttr(self, kloc, idx, dist, cache):
        """
        Example:
            >>> dist = chaospy.J(chaospy.Uniform(), chaospy.Normal(), chaospy.Exponential())
            >>> dist.ttr([[1, 2, 3], [1, 2, 3], [1, 2, 3]]).round(4)
            array([[[0.5   , 0.5   , 0.5   ],
                    [0.    , 0.    , 0.    ],
                    [3.    , 5.    , 7.    ]],
            <BLANKLINE>
                   [[0.0833, 0.0667, 0.0643],
                    [1.    , 2.    , 3.    ],
                    [1.    , 4.    , 9.    ]]])
            >>> d0 = chaospy.Uniform()
            >>> dist = chaospy.J(d0, d0+chaospy.Uniform())
            >>> dist.ttr([1, 1])  # doctest: +IGNORE_EXCEPTION_DETAIL
            Traceback (most recent call last):
                ...
            chaospy.distributions.baseclass.UnsupportedFeature: Joint ...
        """
        if self.stochastic_dependent:
            raise chaospy.UnsupportedFeature(
                "Joint distribution with dependencies not supported.")
        return dist._get_ttr(kloc, idx)

    def __getitem__(self, index):
        """
        Example:
            >>> dist = chaospy.J(chaospy.Uniform(), chaospy.Normal())
            >>> dist[0]
            Uniform()
            >>> dist[1]
            Normal(mu=0, sigma=1)
            >>> dist[:1]
            J(Uniform())
            >>> dist[:2]
            J(Uniform(), Normal(mu=0, sigma=1))
            >>> dist[2]
            Traceback (most recent call last):
                ...
            IndexError: list index out of range
        """
        if isinstance(index, int):
            if index <= -len(self) or index >= len(self):
                raise IndexError("list index out of range")
            return self.get_parameters(index, cache={}, assert_numerical=False)["dist"]
        if isinstance(index, slice):
            start = 0 if index.start is None else index.start
            stop = len(self) if index.stop is None else index.stop
            step = 1 if index.step is None else index.step
            index = range(start, stop, step)
        return J(*[self[idx] for idx in index])

    def _cache(self, idx, cache):
        if idx is None:
            return self
        parameters = self.get_parameters(idx, cache, assert_numerical=False)
        out = parameters["dist"]._get_cache(parameters["idx"], cache)
        if isinstance(out, chaospy.Distribution):
            return self
        return out


# def J(*args, **kwargs):
#     """
#     Joint random variable.

#     Too be deprecated, use `chaospy.J` instead.

#     Args:
#         args (chaospy.Distribution):
#             Distribution to join together.
#     """
#     logger = logging.getLogger(__name__)
#     logger.warning("DepricationWarning: J to be replaced with Joint.")
#     return Joint(*args, **kwargs)



"""
Gauss-Radau formula for numerical estimation of integrals. It requires
:math:`m+1` points and fits all Polynomials to degree :math:`2m`, so it
effectively fits exactly all Polynomials of degree :math:`2m-3`.

Gauss-Radau is defined by having two abscissas to be fixed to the endpoints,
while the others are built around these points. So if a distribution is defined
on the interval ``(a, b)``, then both ``a`` and ``b`` are abscissas in this
scheme. Note though that this does not always possible to achieve in practice,
and an error might be raised.

Example usage
-------------

With increasing order::

    >>> distribution = chaospy.Beta(2, 2, lower=-1, upper=1)
    >>> for order in range(4):  # doctest: +NORMALIZE_WHITESPACE
    ...     X, W = chaospy.generate_quadrature(
    ...         order, distribution, rule="gauss_lobatto")
    ...     print(X.round(2), W.round(2))
    [[-1.]] [1.]
    [[-1.  1.]] [0.5 0.5]
    [[-1.   -0.38  0.38  1.  ]] [0.03 0.47 0.47 0.03]
    [[-1.   -0.69 -0.25  0.25  0.69  1.  ]]
     [0.01 0.15 0.35 0.35 0.15 0.01]

Multivariate samples::

    >>> distribution = chaospy.J(chaospy.Uniform(0, 1), chaospy.Beta(4, 5))
    >>> X, W = chaospy.generate_quadrature(
    ...     2, distribution, rule="gauss_lobatto")
    >>> X.round(3)
    array([[-0.   , -0.   , -0.   , -0.   ,  0.276,  0.276,  0.276,  0.276,
             0.724,  0.724,  0.724,  0.724,  1.   ,  1.   ,  1.   ,  1.   ],
           [ 0.   ,  0.318,  0.605,  1.   ,  0.   ,  0.318,  0.605,  1.   ,
             0.   ,  0.318,  0.605,  1.   ,  0.   ,  0.318,  0.605,  1.   ]])
    >>> W.round(3)
    array([0.001, 0.045, 0.037, 0.   , 0.006, 0.224, 0.184, 0.002, 0.006,
           0.224, 0.184, 0.002, 0.001, 0.045, 0.037, 0.   ])
"""
import numpy
from scipy.linalg import solve_banded, solve

from .recurrence import (
    construct_recurrence_coefficients, coefficients_to_quadrature)
from .combine import combine_quadrature


def quad_gauss_lobatto(
        order,
        dist,
        rule="fejer",
        accuracy=100,
        recurrence_algorithm="",
):
    """
    Generate the abscissas and weights in Gauss-Loboto quadrature.

    Args:
        order (int):
            Quadrature order.
        dist (chaospy.distributions.baseclass.Distribution):
            The distribution weights to be used to create higher order nodes
            from.
        rule (str):
            In the case of ``lanczos`` or ``stieltjes``, defines the
            proxy-integration scheme.
        accuracy (int):
            In the case ``rule`` is used, defines the quadrature order of the
            scheme used. In practice, must be at least as large as ``order``.
        recurrence_algorithm (str):
            Name of the algorithm used to generate abscissas and weights. If
            omitted, ``analytical`` will be tried first, and ``stieltjes`` used
            if that fails.

    Returns:
        (numpy.ndarray, numpy.ndarray):
            abscissas:
                The quadrature points for where to evaluate the model function
                with ``abscissas.shape == (len(dist), N)`` where ``N`` is the
                number of samples.
            weights:
                The quadrature weights with ``weights.shape == (N,)``.

    Example:
        >>> abscissas, weights = quad_gauss_lobatto(
        ...     4, chaospy.Uniform(-1, 1))
        >>> abscissas.round(3)
        array([[-1.   , -0.872, -0.592, -0.209,  0.209,  0.592,  0.872,  1.   ]])
        >>> weights.round(3)
        array([0.018, 0.105, 0.171, 0.206, 0.206, 0.171, 0.105, 0.018])
    """
    assert not rule.startswith("gauss"), "recursive Gaussian quadrature call"
    if order == 0:
        return dist.lower.reshape(1, -1), numpy.array([1.])

    coefficients = construct_recurrence_coefficients(
        2*order-1, dist, rule, accuracy, recurrence_algorithm)
    coefficients = [_lobatto(coeffs, (lo, up))
                    for coeffs, lo, up in zip(coefficients, dist.lower, dist.upper)]
    abscissas, weights = coefficients_to_quadrature(coefficients)

    return combine_quadrature(abscissas, weights)


def _lobatto(coefficients, preassigned):
    """
    Compute the Lobatto nodes and weights with the preassigned value pair.
    Based on the section 7 of the paper

        Some modified matrix eigenvalue problems,
        Gene Golub,
        SIAM Review Vol 15, No. 2, April 1973, pp.318--334,

    and

        http://www.scientificpython.net/pyblog/radau-quadrature

    Args:
        coefficients (numpy.ndarray):
            Three terms recurrence coefficients.
        preassigned (Tuple[float, float]):
            Values that are assume to be fixed.
    """
    alpha = numpy.array(coefficients[0])
    beta = numpy.array(coefficients[1])
    vec_en = numpy.zeros(len(alpha)-1)
    vec_en[-1] = 1
    mat_a1 = numpy.vstack((numpy.sqrt(beta), alpha-preassigned[0]))
    mat_j1 = numpy.vstack((mat_a1[:, 0:-1], mat_a1[0, 1:]))
    mat_a2 = numpy.vstack((numpy.sqrt(beta), alpha - preassigned[1]))
    mat_j2 = numpy.vstack((mat_a2[:, 0:-1], mat_a2[0, 1:]))
    mat_g1 = solve_banded((1, 1), mat_j1, vec_en)
    mat_g2 = solve_banded((1, 1), mat_j2, vec_en)
    mat_c = numpy.array(((1, -mat_g1[-1]), (1, -mat_g2[-1])))
    alpha[-1], beta[-1] = solve(mat_c, preassigned)

    return numpy.array([alpha, beta])



r"""
Archimedean copula definition.

All Copulas can be converted into Cumulative Distribution Functions (CDF) using
the following formula:

.. math::
  \begin{eqnarray}
    F(x_1) &= \frac{C(x_1, 1, 1, ...)}{C(1, 1, 1)} \\
    F(x_2\mid x_1) &= \frac{d/dx_1 C(x_1, 1, 1, ...)}{d/dx_1 C(1, 1, 1, ...)} \\
    F(x_3\mid x_1,x_2) &=
        \frac{d/dx_1 d/dx_2 C(x_1, x_2, x_3, 1, ...) }{
              d/dx_1 d/dx_2 C(x_1, x_2, 1, 1, ...)}
  \end{eqnarray}

And thereby density functions:

.. math::
  \begin{eqnarray}
    P(x_1) &= \frac{d/dx_1 C(x_1, 1, 1, ...)}{C(1, 1, 1)} \\
    P(x_2\mid x_1) &= \frac{d/dx_1 d/dx_2 C(x_1, 1, 1, ...)}{d/dx_1 C(1, 1, 1, ...)} \\
    P(x_3\mid x_1,x_2) &=
        \frac{d/dx_1 d/dx_2 d/dx_3 C(x_1, x_2, x_3, 1, ...) }{
              d/dx_1 d/dx_2 C(x_1, x_2, 1, 1, ...)}
  \end{eqnarray}

The general definition of an Archimedean is:

.. math::
    C(x_1, x_2, x_3, ...) = iphi( phi(x_1) + phi(x_2) + phi(x_3) + ... )

Applying the Copula to CDF formula on the Archimedean, we need various partial
derivatives:

.. math::
  \begin{eqnarray}
    C &= iphi(phi(x_1)+phi(x_2)+phi(x_3)+...) \\
    d/dx_1 C &= iphi'( phi(x_1)+phi(x_2)+phi(x_3)+...) phi'(x_1) \\
    d/dx_1 d/dx_2 C &= iphi''(phi(x_1)+phi(x_2)+phi(x_3)+...) phi'(x_1) phi'(x_2) \\
    d/dx_1 d/dx_2 d/dx_3 C &= iphi'''(phi(x_1)+phi(x_2)+phi(x_3)+...) phi'(x_1) phi'(x_2) phi'(x_3) \\
    & \vdots \\
    d/dx_1 \cdots d/dx_n C &= iphi^{(n)}(phi(x_1)+phi(x_2)+phi(x_3)+...) phi'(x_1) \cdots phi'(x_n)
  \end{eqnarray}

As a helper function, we introduce sigma:

.. math::
    \sigma(r, n, \theta) = \prod_{d=0}^{n-1}(-1/\theta-d)*r^{-1/theta-n}

It has the convenient property:

.. math::
    \sigma'(r, n, \theta) = \prod_{d=0}^{n}(-1/\theta-d)*r^{-1/theta-n-1} = \sigma(r, n+1, \theta)
"""
import numpy
import chaospy

from ..baseclass import Distribution


class Archimedean(Distribution):

    def __init__(self, length, theta=1., rotation=None):
        if rotation is not None:
            assert length == len(rotation)
        dependencies, _, rotation = chaospy.declare_dependencies(
            self,
            parameters=dict(theta=theta),
            rotation=rotation,
            dependency_type="accumulate",
            length=length,
        )
        super(Archimedean, self).__init__(
            parameters=dict(theta=float(theta)),
            dependencies=dependencies,
            rotation=rotation,
            repr_args=[length, "theta=%s" % theta],
        )

    def get_parameters(self, idx, cache, assert_numerical=True):
        parameters = super(Archimedean, self).get_parameters(
            idx, cache, assert_numerical=assert_numerical)
        theta = parameters["theta"]

        if idx is None:
            return dict(theta=theta, cache=cache)
        return dict(idx=idx, theta=theta, cache=cache)

    def _lower(self, idx, theta, cache):
        return 0.

    def _upper(self, idx, theta, cache):
        return 1.

    def _ppf(self, qloc, idx, theta, cache):
        raise chaospy.UnsupportedFeature("Copula not supported.")

    def _cdf(self, xloc, idx, theta, cache):
        dim = self._rotation.index(idx)
        conditions = [self[dim_]._get_cache(0, cache, get=0)
                      for dim_ in self._rotation[:dim]]
        assert not any([isinstance(condition, chaospy.Distribution)
                        for condition in conditions])
        ones = numpy.ones((len(self)-len(conditions), xloc.shape[-1]))
        xloc1 = numpy.vstack(conditions+[xloc, ones[:-1]])
        xloc2 = numpy.vstack(conditions+[ones])
        out1 = self._copula(xloc1, theta, order=idx)
        out2 = self._copula(xloc2, theta, order=idx)
        out = numpy.where(out2, out1, 0)/numpy.where(out2, out2, 1)
        return out

    def _pdf(self, xloc, idx, theta, cache):
        dim = self._rotation.index(idx)
        conditions = [self[dim_]._get_cache(0, cache, get=0)
                      for dim_ in self._rotation[:dim]]
        assert not any([isinstance(condition, chaospy.Distribution)
                        for condition in conditions])
        ones = numpy.ones((len(self)-len(conditions), xloc.shape[-1]))
        xloc1 = numpy.vstack(conditions+[xloc, ones[:-1]])
        xloc2 = numpy.vstack(conditions+[ones])
        out1 = self._copula(xloc1, theta, order=idx+1)
        out2 = self._copula(xloc2, theta, order=idx)
        out = numpy.where(out2, out1, 0)/numpy.where(out2, out2, 1)
        return out

    def _copula(self, x_loc, theta, order=0):
        assert isinstance(x_loc, numpy.ndarray)
        assert len(x_loc) == len(self)
        theta = float(theta)
        order = int(order)

        out = numpy.sum(self._phi(x_loc, theta), 0)
        out = self._inverse_phi(out, theta, order)
        if order:
            out *= numpy.where(
                out, numpy.prod(self._delta_phi(x_loc[:order], theta), 0), 0)
        else:
            out = numpy.clip(out, 0, 1)
        return out

    @staticmethod
    def _sigma(u_loc, theta, order):
        out = 1.
        for dim in range(order):
            out *= (-1/theta-dim)
        out = out*u_loc**(-1/theta-order)
        return out



"""
In mathematics, a `low-discrepancy sequence`_ is a sequence with the property
that for all values of N, its subsequence x1, ..., xN has a low discrepancy.

Roughly speaking, the discrepancy of a sequence is low if the proportion of
points in the sequence falling into an arbitrary set B is close to proportional
to the measure of B, as would happen on average (but not for particular
samples) in the case of an equi-distributed sequence. Specific definitions of
discrepancy differ regarding the choice of B (hyperspheres, hypercubes, etc.)
and how the discrepancy for every B is computed (usually normalized) and
combined (usually by taking the worst value).

Low-discrepancy sequences are also called quasi-random or sub-random sequences,
due to their common use as a replacement of uniformly distributed random
numbers. The "quasi" modifier is used to denote more clearly that the values of
a low-discrepancy sequence are neither random nor pseudorandom, but such
sequences share some properties of random variables and in certain applications
such as the quasi-Monte Carlo method their lower discrepancy is an important
advantage.

.. low-discrepancy sequence: https://en.wikipedia.org/wiki/Low-discrepancy_sequence
"""
from .chebyshev import create_chebyshev_samples, create_nested_chebyshev_samples
from .grid import create_grid_samples, create_nested_grid_samples
from .halton import create_halton_samples
from .hammersley import create_hammersley_samples
from .sobol import create_sobol_samples
from .korobov import create_korobov_samples



"""Pareto type 2 distribution."""
from ..baseclass import SimpleDistribution, ShiftScaleDistribution


class pareto2(SimpleDistribution):
    """Pareto type 2 distribution."""

    def __init__(self, c):
        super(pareto2, self).__init__(dict(c=c))

    def _pdf(self, x, c):
        return c*1.0/(1.0+x)**(c+1.0)

    def _cdf(self, x, c):
        return 1.0-1.0/(1.0+x)**c

    def _ppf(self, q, c):
        return pow(1.0-q, -1.0/c)-1

    def _lower(self, c):
        return 0.0

    def _upper(self, c):
        return pow(1e-12, -1./c)-1


class Pareto2(ShiftScaleDistribution):
    """
    Pareto type 2 distribution.

    Also known as Lomax distribution (for loc=0).

    Lower threshold at loc and survival: (1+x)^-shape.

    Args:
        shape (float, Distribution):
            Shape parameter
        scale (float, Distribution):
            Scaling parameter
        loc (float, Distribution):
            Location parameter

    Examples:
        >>> distribution = chaospy.Pareto2(15)
        >>> distribution
        Pareto2(15)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([0.   , 0.015, 0.035, 0.063, 0.113, 5.31 ])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([15.   , 11.823,  8.699,  5.644,  2.695,  0.   ])
        >>> distribution.sample(4).round(3)
        array([0.073, 0.008, 0.222, 0.045])

    """

    def __init__(self, shape=1, scale=1, shift=0):
        super(Pareto2, self).__init__(
            dist=pareto2(shape),
            scale=scale,
            shift=shift,
            repr_args=[shape],
        )



# -*- coding: utf-8 -*-
"""
`OpenTURNS`_ is an open source initiative for the treatment of uncertainties,
risks and statistics in a structured industrial approach. It is a mature
toolkit with a lot of functionality beyond what ``chaospy`` delivers. If one
wants to combine the strength of the two projects, it is possible to interpret
distributions in `OpenTURNS`_ as ``chaospy`` distributions or vise versa.

To interpret a ``chaospy`` distribution as an `OpenTURNS`_ distribution, see
the `OpenTURNS distribution wrapper`_.

To interpret an `OpenTURNS`_ distribution as a ``chaospy`` distribution, it is
possible to use the :class:`OpenTURNSDist` wrapper. For example to interpret
a simple univariate Gaussian random variable::

    >>> import openturns
    >>> distribution = chaospy.OpenTURNSDist(openturns.Normal(0, 1))
    >>> distribution
    OpenTURNSDist([openturns.Normal(mu = 0, sigma = 1)])

This distribution then behaves as a normal ``chaospy`` distribution::

    >>> distribution.pdf([-1, 0, 1]).round(4)
    array([0.242 , 0.3989, 0.242 ])
    >>> distribution.mom([0, 1, 2])
    array([1., 0., 1.])

The wrapper also supports multivariate distributions::

    >>> composed_distribution = openturns.ComposedDistribution([
    ...     openturns.Triangular(-1, 0, 1), openturns.Uniform(-1, 1)])
    >>> mv_distribution = chaospy.OpenTURNSDist(composed_distribution)
    >>> mv_distribution  # doctest: +NORMALIZE_WHITESPACE
    OpenTURNSDist([openturns.Triangular(a = -1, m = 0, b = 1),
                   openturns.Uniform(a = -1, b = 1)])
    >>> mv_distribution.sample(4).round(4)
    array([[ 0.1676, -0.5204,  0.6847, -0.018 ],
           [ 0.7449, -0.5753, -0.9186, -0.2056]])

As a shorthand, it is also possible to construct multivariate distributions
form lists::

    >>> mv_distribution = chaospy.OpenTURNSDist([
    ...     openturns.Triangular(-1, 0, 1), openturns.Uniform(-1, 1)])
    >>> mv_distribution  # doctest: +NORMALIZE_WHITESPACE
    OpenTURNSDist([openturns.Triangular(a = -1, m = 0, b = 1),
                   openturns.Uniform(a = -1, b = 1)])

Though multivariate distributions are supported, dependencies are not::

    >>> correlation = openturns.CorrelationMatrix(2)
    >>> correlation[0, 1] = 0.3
    >>> copula = openturns.NormalCopula(correlation)
    >>> dependent_distribution = openturns.ComposedDistribution(
    ...     [openturns.Normal(), openturns.Uniform()], copula)
    >>> chaospy.OpenTURNSDist(dependent_distribution)  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
        ...
    chaospy.StochasticallyDependentError: Stochastically dependent OpenTURNS distribution unsupported

.. _OpenTURNS distribution wrapper: http://openturns.github.io/\
openturns/latest/user_manual/_generated/openturns.ChaospyDistribution.html
"""
try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable
import numpy
import chaospy

from ..distributions import SimpleDistribution, J


class openturns_dist(SimpleDistribution):
    """One dimensional OpenTURNS distribution."""

    def __init__(self, distribution):
        """
        Args:
            distribution (openturns.Distribution):
                Distribution created in OpenTURNS.
        """
        assert distribution.isContinuous(), (
            "Only continuous distributions are supported")
        super(openturns_dist, self).__init__()
        self.distribution = distribution

    def _pdf(self, x_loc):
        return numpy.array(self.distribution.computePDF(
            numpy.atleast_2d(x_loc).T).asPoint())

    def _cdf(self, x_loc):
        return numpy.array(self.distribution.computeCDF(
            numpy.atleast_2d(x_loc).T).asPoint())

    def _ppf(self, q_loc):
        return numpy.array(
            self.distribution.computeQuantile(q_loc).asPoint())

    def _lower(self):
        return self.distribution.getRange().getLowerBound()[0]

    def _upper(self):
        return self.distribution.getRange().getUpperBound()[0]

    def _mom(self, k_loc):
        return self.distribution.getMoment(int(k_loc))[0]

    def __str__(self):
        return "openturns."+str(self.distribution)

    def __repr__(self):
        return str(self)


class OpenTURNSDist(J):
    """
    OpenTURNS distribution constructor.

    Args:
        distribution (openturns.Distribution):
            Distribution created in OpenTURNS.

    Examples:
        >>> from openturns import ComposedDistribution, Normal
        >>> ot_distribution = ComposedDistribution([Normal(), Normal()])
        >>> distribution = chaospy.OpenTURNSDist(ot_distribution)
        >>> distribution  # doctest: +NORMALIZE_WHITESPACE
        OpenTURNSDist([openturns.Normal(mu = 0, sigma = 1),
                       openturns.Normal(mu = 0, sigma = 1)])
    """

    def __init__(self, distribution):
        from openturns import ComposedDistribution, ContinuousDistribution
        if isinstance(distribution, ComposedDistribution):
            if not distribution.hasIndependentCopula():
                raise chaospy.StochasticallyDependentError(
                    "Stochastically dependent "
                    "OpenTURNS distribution unsupported")
            distributions = [
                openturns_dist(dist)
                for dist in distribution.getDistributionCollection()
            ]
        elif isinstance(distribution, ContinuousDistribution):
            distributions = [openturns_dist(distribution)]
        else:
            assert isinstance(distribution, Iterable) and all([
                isinstance(dist, ContinuousDistribution)
                for dist in distribution
            ]), "Only (iterable of) continuous OpenTURNS distributions supported"
            distributions = [openturns_dist(dist) for dist in distribution]
        super(OpenTURNSDist, self).__init__(*distributions)
        self._repr_args = [distributions]



"""
Generate samples from a regular grid.

Example usage
-------------

Basic usage::

    >>> distribution = chaospy.Uniform(0, 1)
    >>> samples = distribution.sample(2, rule="grid")
    >>> samples.round(4)
    array([0.3333, 0.6667])
    >>> samples = distribution.sample(5, rule="grid")
    >>> samples.round(4)
    array([0.1667, 0.3333, 0.5   , 0.6667, 0.8333])

Certain orders are nested::

    >>> samples = distribution.sample(3, rule="grid")
    >>> samples.round(4)
    array([0.25, 0.5 , 0.75])
    >>> samples = distribution.sample(7, rule="grid")
    >>> samples.round(4)
    array([0.125, 0.25 , 0.375, 0.5  , 0.625, 0.75 , 0.875])

Create nested samples directly with the dedicated function::

    >>> samples = distribution.sample(2, rule="nested_grid")
    >>> samples.round(4)
    array([0.25, 0.5 , 0.75])
    >>> samples = distribution.sample(3, rule="nested_grid")
    >>> samples.round(4)
    array([0.125, 0.25 , 0.375, 0.5  , 0.625, 0.75 , 0.875])

Multivariate usage::

    >>> distribution = chaospy.J(chaospy.Uniform(0, 1), chaospy.Uniform(0, 1))
    >>> samples = distribution.sample(2, rule="grid")
    >>> samples.round(4)
    array([[0.3333, 0.3333, 0.6667, 0.6667],
           [0.3333, 0.6667, 0.3333, 0.6667]])

"""
import numpy
import chaospy


def create_grid_samples(order, dim=1):
    """
    Create samples from a regular grid.

    Args:
        order (int):
            The order of the grid. Defines the number of samples.
        dim (int):
            The number of dimensions in the grid

    Returns (numpy.ndarray):
        Regular grid with ``shape == (dim, order)``.
    """
    x_data = numpy.arange(1, order+1)/(order+1.)
    x_data = chaospy.quadrature.combine([x_data]*dim)
    return x_data.T


def create_nested_grid_samples(order, dim=1):
    """
    Create samples from a nested grid.

    Args:
        order (int):
            The order of the grid. Defines the number of samples.
        dim (int):
            The number of dimensions in the grid

    Returns (numpy.ndarray):
        Regular grid with ``shape == (dim, 2**order-1)``.
    """
    return create_grid_samples(order=2**order-1, dim=dim)



"""
Generates samples from the `Sobol sequence`_.

Sobol sequences (also called LP_T sequences or (t, s) sequences in base 2) are
an example of quasi-random low-discrepancy sequences. They were first
introduced by the Russian mathematician Ilya M. Sobol in 1967.

These sequences use a base of two to form successively finer uniform partitions
of the unit interval and then reorder the coordinates in each dimension.

Example usage
-------------

Standard usage::

    >>> distribution = chaospy.J(chaospy.Uniform(0, 1), chaospy.Uniform(0, 1))
    >>> samples = distribution.sample(3, rule="sobol")
    >>> samples.round(4)
    array([[0.5 , 0.75, 0.25],
           [0.5 , 0.25, 0.75]])
    >>> samples = distribution.sample(4, rule="sobol")
    >>> samples.round(4)
    array([[0.5  , 0.75 , 0.25 , 0.375],
           [0.5  , 0.25 , 0.75 , 0.375]])

Papers::

    Antonov, Saleev,
    USSR Computational Mathematics and Mathematical Physics,
    Volume 19, 1980, pages 252 - 256.

    Paul Bratley, Bennett Fox,
    Algorithm 659:
    Implementing Sobol's Quasirandom Sequence Generator,
    ACM Transactions on Mathematical Software,
    Volume 14, Number 1, pages 88-100, 1988.

    Bennett Fox,
    Algorithm 647:
    Implementation and Relative Efficiency of Quasirandom
    Sequence Generators,
    ACM Transactions on Mathematical Software,
    Volume 12, Number 4, pages 362-376, 1986.

    Ilya Sobol,
    USSR Computational Mathematics and Mathematical Physics,
    Volume 16, pages 236-242, 1977.

    Ilya Sobol, Levitan,
    The Production of Points Uniformly Distributed in a Multidimensional
    Cube (in Russian),
    Preprint IPM Akad. Nauk SSSR,
    Number 40, Moscow 1976.

.. _Sobel sequence: https://en.wikipedia.org/wiki/Sobol_sequence
"""
import math

import numpy

DIM_MAX = 40
LOG_MAX = 30

SOURCE_SAMPLES = numpy.zeros((DIM_MAX, LOG_MAX), dtype=int)
SOURCE_SAMPLES[0:40, 0] = 1
SOURCE_SAMPLES[2:40, 1] = (
    1, 3, 1, 3, 1, 3, 3, 1, 3, 1, 3, 1, 3, 1, 1, 3, 1, 3, 1, 3, 1, 3, 3, 1, 3,
    1, 3, 1, 3, 1, 1, 3, 1, 3, 1, 3, 1, 3
)
SOURCE_SAMPLES[3:40, 2] = (
    7, 5, 1, 3, 3, 7, 5, 5, 7, 7, 1, 3, 3, 7, 5, 1, 1, 5, 3, 3, 1, 7, 5, 1, 3,
    3, 7, 5, 1, 1, 5, 7, 7, 5, 1, 3, 3
)
SOURCE_SAMPLES[5:40, 3] = (
    1, 7, 9, 13, 11, 1, 3, 7, 9, 5, 13, 13, 11, 3, 15, 5, 3, 15, 7, 9, 13, 9,
    1, 11, 7, 5, 15, 1, 15, 11, 5, 3, 1, 7, 9
)
SOURCE_SAMPLES[7:40, 4] = (
    9, 3, 27, 15, 29, 21, 23, 19, 11, 25, 7, 13, 17, 1, 25, 29, 3, 31, 11, 5,
    23, 27, 19, 21, 5, 1, 17, 13, 7, 15, 9, 31, 9
)
SOURCE_SAMPLES[13:40, 5] = (
    37, 33, 7, 5, 11, 39, 63, 27, 17, 15, 23, 29, 3, 21, 13, 31, 25, 9, 49, 33,
    19, 29, 11, 19, 27, 15, 25
)
SOURCE_SAMPLES[19:40, 6] = (
    13, 33, 115, 41, 79, 17, 29, 119, 75, 73, 105, 7, 59, 65, 21, 3, 113, 61,
    89, 45, 107
)
SOURCE_SAMPLES[37:40, 7] = (7, 23, 39)
POLY = (
    1, 3, 7, 11, 13, 19, 25, 37, 59, 47, 61, 55, 41, 67, 97, 91, 109, 103,
    115, 131, 193, 137, 145, 143, 241, 157, 185, 167, 229, 171, 213, 191,
    253, 203, 211, 239, 247, 285, 369, 299
)


def create_sobol_samples(order, dim, seed=1):
    """
    Args:
        order (int):
            Number of unique samples to generate.
        dim (int):
            Number of spacial dimensions. Must satisfy ``0 < dim < 41``.
        seed (int):
            Starting seed. Non-positive values are treated as 1. If omitted,
            consecutive samples are used.

    Returns:
        (numpy.ndarray):
            Quasi-random vector with ``shape == (dim, order)``.
    """
    assert 0 < dim < DIM_MAX, "dim in [1, 40]"

    # Initialize row 1 of V.
    samples = SOURCE_SAMPLES.copy()
    maxcol = int(math.log(2**LOG_MAX-1, 2))+1
    samples[0, 0:maxcol] = 1

    # Initialize the remaining rows of V.
    for idx in range(1, dim):

        # The bits of the integer POLY(I) gives the form of polynomial:
        degree = int(math.log(POLY[idx], 2))

        #Expand this bit pattern to separate components:
        includ = numpy.array([val == "1" for val in bin(POLY[idx])[-degree:]])

        #Calculate the remaining elements of row I as explained
        #in Bratley and Fox, section 2.
        for idy in range(degree+1, maxcol+1):
            newv = samples[idx, idy-degree-1].item()
            base = 1
            for idz in range(1, degree+1):
                base *= 2
                if includ[idz-1]:
                    newv = newv ^ base * samples[idx, idy-idz-1].item()
            samples[idx, idy-1] = newv

    samples = samples[:dim]

    # Multiply columns of V by appropriate power of 2.
    samples *= 2**(numpy.arange(maxcol, 0, -1, dtype=int))

    #RECIPD is 1/(common denominator of the elements in V).
    recipd = 0.5**(maxcol+1)
    lastq = numpy.zeros(dim, dtype=int)

    seed = int(seed) if seed > 1 else 1

    for seed_ in range(seed):
        lowbit = len(bin(seed_)[2:].split("0")[-1])
        lastq[:] = lastq ^ samples[:, lowbit]

    #Calculate the new components of QUASI.
    quasi = numpy.empty((dim, order))
    for idx in range(order):
        lowbit = len(bin(seed+idx)[2:].split("0")[-1])
        quasi[:, idx] = lastq * recipd
        lastq[:] = lastq ^ samples[:, lowbit]

    return quasi



"""Folded normal distribution."""
import numpy
from scipy import special

from ..baseclass import SimpleDistribution, ShiftScaleDistribution


class folded_normal(SimpleDistribution):
    """Folded normal distribution."""

    def __init__(self, c=1):
        super(folded_normal, self).__init__(dict(c=c))

    def _pdf(self, x, c):
        return numpy.sqrt(2.0/numpy.pi)*numpy.cosh(c*x)*numpy.exp(-(x*x+c*c)/2.0)

    def _cdf(self, x, c):
        return special.ndtr(x-c)+special.ndtr(x+c)-1.0

    def _lower(self, c):
        return 0.

    def _upper(self, c):
        return 8+c


class FoldedNormal(ShiftScaleDistribution):
    """
    Folded normal distribution.

    Args:
        mu (float, Distribution):
            Location parameter in normal distribution.
        scale (float, Distribution):
            Scaling parameter (in both normal and fold).
        shift (float, Distribution):
            Location of fold.

    Examples:
        >>> distribution = chaospy.FoldedNormal(1.5)
        >>> distribution
        FoldedNormal(1.5)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([0.   , 0.706, 1.254, 1.755, 2.342, 9.5  ])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0.259, 0.326, 0.396, 0.388, 0.28 , 0.   ])
        >>> distribution.sample(4).round(3)
        array([1.906, 2.225, 1.638, 2.701])

    """

    def __init__(self, mu=0, scale=1, shift=0):
        super(FoldedNormal, self).__init__(
            dist=folded_normal(mu-shift),
            scale=scale,
            shift=shift,
            repr_args=[mu],
        )



"""Anglit distribution."""
import numpy

from ..baseclass import SimpleDistribution, ShiftScaleDistribution


class anglit(SimpleDistribution):
    """Anglit distribution."""

    def __init__(self):
        super(anglit, self).__init__()

    def _pdf(self, x):
        return numpy.cos(2*x)

    def _cdf(self, x):
        return numpy.sin(x+numpy.pi/4)**2.0

    def _ppf(self, q):
        return (numpy.arcsin(numpy.sqrt(q))-numpy.pi/4)

    def _lower(self):
        return -numpy.pi/4

    def _upper(self):
        return numpy.pi/4


class Anglit(ShiftScaleDistribution):
    """
    Anglit distribution.

    Args:
        loc (float, Distribution):
            Location parameter
        scale (float, Distribution):
            Scaling parameter

    Examples:
        >>> distribution = chaospy.Anglit()
        >>> distribution
        Anglit()
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([-0.785, -0.322, -0.101,  0.101,  0.322,  0.785])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0.  , 0.8 , 0.98, 0.98, 0.8 , 0.  ])
        >>> distribution.sample(4).round(3)
        array([ 0.156, -0.439,  0.561, -0.018])

    """

    def __init__(self, scale=1, shift=0):
        super(Anglit, self).__init__(dist=anglit(), scale=scale, shift=shift)



r"""
The Gauss-Legendre quadrature rule is properly supported by in :ref:`gaussian`.
However, as Gauss-Legendre is a special case where the weight function is
constant, it can in principle be used to integrate any weighting function. In
other words, this is the same Gauss-Legendre integration rule, but only in the
context of uniform distribution as weight function. Normalization of the
weights will be used to achieve the general integration form.

It is also worth noting that this specific implementation of Gauss-Legendre is
faster to compute than the general version in :ref:`gaussian`.

Example usage
-------------

The first few orders::

    >>> distribution = chaospy.Uniform(0, 1)
    >>> for order in [0, 1, 2, 3]:
    ...     abscissas, weights = chaospy.generate_quadrature(
    ...         order, distribution, rule="gauss_legendre")
    ...     print(order, abscissas.round(3), weights.round(3))
    0 [[0.5]] [1.]
    1 [[0.211 0.789]] [0.5 0.5]
    2 [[0.113 0.5   0.887]] [0.278 0.444 0.278]
    3 [[0.069 0.33  0.67  0.931]] [0.174 0.326 0.326 0.174]

Using an alternative distribution::

    >>> distribution = chaospy.Beta(2, 4)
    >>> for order in [0, 1, 2, 3]:
    ...     abscissas, weights = chaospy.generate_quadrature(
    ...         order, distribution, rule="gauss_legendre")
    ...     print(order, abscissas.round(3), weights.round(3))
    0 [[0.5]] [1.]
    1 [[0.211 0.789]] [0.933 0.067]
    2 [[0.113 0.5   0.887]] [0.437 0.556 0.007]
    3 [[0.069 0.33  0.67  0.931]] [0.195 0.647 0.157 0.001]

The abscissas stays the same, but the weights are re-adjusted for the new
weight function.
"""
import numpy

from .recurrence import (
    construct_recurrence_coefficients, coefficients_to_quadrature)
from .combine import combine_quadrature


def quad_gauss_legendre(
        order,
        domain=(0, 1),
        rule="fejer",
        accuracy=100,
        recurrence_algorithm="",
):
    r"""
    Generate the quadrature nodes and weights in Gauss-Legendre quadrature.

    Note that this rule exists to allow for integrating functions with weight
    functions without actually adding the quadrature. Like:

    .. math:
        \int_a^b p(x) f(x) dx \approx \sum_i p(X_i) f(X_i) W_i

    instead of the more traditional:

    .. math:
        \int_a^b p(x) f(x) dx \approx \sum_i f(X_i) W_i

    To get the behavior where the weight function is taken into consideration,
    use :func:`~chaospy.quadrature.gaussian.quad_gaussian`.

    Args:
        order (int, numpy.ndarray):
            Quadrature order.
        domain (chaospy.distributions.baseclass.Distribution, numpy.ndarray):
            Either distribution or bounding of interval to integrate over.
        rule (str):
            In the case of ``lanczos`` or ``stieltjes``, defines the
            proxy-integration scheme.
        accuracy (int):
            In the case ``rule`` is used, defines the quadrature order of the
            scheme used. In practice, must be at least as large as ``order``.
        recurrence_algorithm (str):
            Name of the algorithm used to generate abscissas and weights. If
            omitted, ``analytical`` will be tried first, and ``stieltjes`` used
            if that fails.

    Returns:
        (numpy.ndarray, numpy.ndarray):
            abscissas:
                The quadrature points for where to evaluate the model function
                with ``abscissas.shape == (len(dist), N)`` where ``N`` is the
                number of samples.
            weights:
                The quadrature weights with ``weights.shape == (N,)``.

    Example:
        >>> abscissas, weights = quad_gauss_legendre(3)
        >>> abscissas.round(4)
        array([[0.0694, 0.33  , 0.67  , 0.9306]])
        >>> weights.round(4)
        array([0.1739, 0.3261, 0.3261, 0.1739])
    """
    from ..distributions.baseclass import Distribution
    from ..distributions.collection import Uniform
    if isinstance(domain, Distribution):
        abscissas, weights = quad_gauss_legendre(
            order, (domain.lower, domain.upper),
            rule, accuracy, recurrence_algorithm)

        pdf = domain.pdf(abscissas)
        if len(domain) > 1:
            weights = (weights.T*pdf).T
        else:
            weights *= pdf.flatten()
        weights /= numpy.sum(weights)
        return abscissas, weights

    order = numpy.asarray(order, dtype=int).flatten()
    lower, upper = numpy.array(domain)
    lower = numpy.asarray(lower).flatten()
    upper = numpy.asarray(upper).flatten()

    dim = max(lower.size, upper.size, order.size)
    order = numpy.ones(dim, dtype=int)*order
    lower = numpy.ones(dim)*lower
    upper = numpy.ones(dim)*upper

    coefficients = construct_recurrence_coefficients(
        numpy.max(order), Uniform(0, 1), rule, accuracy, recurrence_algorithm)

    abscissas, weights = zip(*[coefficients_to_quadrature(
        coefficients[:order_+1]) for order_ in order])
    abscissas = list(numpy.asarray(abscissas).reshape(dim, -1))
    weights = list(numpy.asarray(weights).reshape(dim, -1))

    return combine_quadrature(abscissas, weights, (lower, upper))



r"""
Monte Carlo simulation is by nature a very slow converging method.  The error
in convergence is proportional to :math:`1/\sqrt{K}` where :math:`K` is the
number of samples.  It is somewhat better with variance reduction techniques
that often reaches errors proportional to :math:`1/K`. For a full overview of
the convergence rate of the various methods, see for example the excellent book
`handbook of Monte Carlo methods` by Kroese, Taimre and Botev. However as the
number of dimensions grows, Monte Carlo convergence rate stays the same, making
it immune to the curse of dimensionality.

Generating random samples can be done from the distribution instance method
``sample`` as discussed in the :ref:`tutorial`. For example, to generate nodes
from the Korobov lattice::

    >>> distribution = chaospy.Iid(chaospy.Beta(2, 2), 2)
    >>> samples = distribution.sample(4, rule="korobov")
    >>> samples.round(4)
    array([[0.2871, 0.4329, 0.5671, 0.7129],
           [0.4329, 0.7129, 0.2871, 0.5671]])

.. _handbook of Monte Carlo methods: https://onlinelibrary.wiley.com/doi/book/10.1002/9781118014967
"""
from .generator import generate_samples

from .sequences import *  # pylint: disable=wildcard-import
from .latin_hypercube import create_latin_hypercube_samples
from .antithetic import create_antithetic_variates



"""Power log-Normal probability distribution."""
import numpy
from scipy import special
import chaospy

from ..baseclass import SimpleDistribution, ShiftScaleDistribution


class power_log_normal(SimpleDistribution):
    """Power log-Normal probability distribution."""

    def __init__(self, c, s):
        super(power_log_normal, self).__init__(dict(c=c, s=s))

    def _pdf(self, x, c, s):
        norm = (2*numpy.pi)**-.5*numpy.exp(-(numpy.log(x)/s)**2/2.)
        out = c/(x*s)*norm*pow(special.ndtr(-numpy.log(x)/s), c*1.-1.)
        out = numpy.where(x == 0, 0, out)
        return out

    def _cdf(self, x, c, s):
        return 1.-pow(special.ndtr(-numpy.log(x)/s), c*1.)

    def _ppf(self, q, c, s):
        return numpy.exp(-s*special.ndtri(pow(1.-q, 1./c)))

    def _lower(self, c, s):
        return 0.

    def _upper(self, c, s):
        return numpy.exp(-s*special.ndtri(pow(1e-12, 1./c)))


class PowerLogNormal(ShiftScaleDistribution):
    """
    Power log-normal distribution

    Args:
        shape (float, Distribution):
            Shape parameter
        mu (float, Distribution):
            Mean in the normal distribution.  Overlaps with scale by
            mu=log(scale)
        sigma (float, Distribution):
            Standard deviation of the normal distribution.
        shift (float, Distribution):
            Location parameter
        scale (float, Distribution):
            Scaling parameter. Overlap with mu in scale=e**mu

    Examples:
        >>> distribution = chaospy.PowerLogNormal(1.5)
        >>> distribution
        PowerLogNormal(1.5)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([  0.   ,   0.337,   0.573,   0.898,   1.502, 273.691])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0.   , 0.912, 0.755, 0.488, 0.214, 0.   ])
        >>> distribution.sample(4).round(3)
        array([1.017, 0.242, 3.01 , 0.69 ])

    """
    def __init__(self, shape=1, mu=0, sigma=1, scale=1, shift=0):
        dist = ShiftScaleDistribution(
            dist=power_log_normal(shape, sigma), scale=numpy.e**mu)
        super(PowerLogNormal, self).__init__(
            dist=dist,
            scale=scale,
            shift=shift,
            repr_args=[shape]+chaospy.format_repr_kwargs(mu=(mu, 0))+
                              chaospy.format_repr_kwargs(sigma=(sigma, 1)),
        )



"""Gamma distribution."""
import numpy
from scipy import special

from ..baseclass import SimpleDistribution, ShiftScaleDistribution


class gamma(SimpleDistribution):

    def __init__(self, a=1):
        super(gamma, self).__init__(dict(a=a))

    def _pdf(self, x, a):
        return x**(a-1)*numpy.e**(-x)/special.gamma(a)

    def _cdf(self, x, a):
        return special.gammainc(a, x)

    def _ppf(self, q, a):
        return special.gammaincinv(a, q)

    def _mom(self, k, a):
        return special.gamma(a+k)/special.gamma(a)

    def _ttr(self, n, a):
        return 2.*n+a, n*n+n*(a-1)

    def _lower(self, a):
        return 0.

    def _upper(self, a):
        return special.gammaincinv(a, 1-1e-12)


class Gamma(ShiftScaleDistribution):
    """
    Gamma distribution.

    Also an Erlang distribution when shape=k and scale=1./lamb.

    Args:
        shape (float, Distribution):
            Shape parameter. a>0.
        scale (float, Distribution):
            Scale parameter. scale!=0
        shift (float, Distribution):
            Location of the lower bound.

    Examples:
        >>> distribution = chaospy.Gamma(3, scale=0.5)
        >>> distribution
        Gamma(3, scale=0.5)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([ 0.   ,  0.768,  1.143,  1.553,  2.14 , 17.026])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0.   , 0.508, 0.531, 0.432, 0.254, 0.   ])
        >>> distribution.sample(4).round(3)
        array([1.683, 0.587, 3.152, 1.301])
        >>> distribution.mom(1).round(3)
        1.5
        >>> distribution.ttr([0, 1, 2, 3]).round(3)
        array([[1.5 , 2.5 , 3.5 , 4.5 ],
               [0.  , 0.75, 2.  , 3.75]])

    """

    def __init__(self, shape=1, scale=1, shift=0):
        super(Gamma, self).__init__(
            dist=gamma(shape),
            scale=scale,
            shift=shift,
            repr_args=[shape],
        )


class Exponential(ShiftScaleDistribution):
    R"""
    Exponential Probability Distribution

    Args:
        scale (float, Distribution):
            Scale parameter. scale!=0
        shift (float, Distribution):
            Location of the lower bound.

    Examples;:
        >>> distribution = chaospy.Exponential()
        >>> distribution
        Exponential()
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([ 0.   ,  0.223,  0.511,  0.916,  1.609, 27.631])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([1. , 0.8, 0.6, 0.4, 0.2, 0. ])
        >>> distribution.sample(4).round(3)
        array([1.06 , 0.122, 3.001, 0.658])
        >>> distribution.mom(1).round(3)
        1.0
        >>> distribution.ttr([1, 2, 3]).round(3)
        array([[3., 5., 7.],
               [1., 4., 9.]])

    """

    def __init__(self, scale=1, shift=0):
        super(Exponential, self).__init__(
            dist=gamma(1),
            scale=scale,
            shift=shift,
            repr_args=[],
        )



r"""Collection of polynomial expansion constructors."""
from .frontend import generate_expansion
from .three_terms_recurrence import orth_ttr
from .lagrange import lagrange_polynomial
from .gram_schmidt import orth_gs
from .cholesky import orth_chol



"""Levy distribution."""
import numpy
from scipy import special

from ..baseclass import SimpleDistribution, ShiftScaleDistribution


class levy(SimpleDistribution):
    """Levy distribution."""

    def __init__(self):
        super(levy, self).__init__()

    def _pdf(self, x):
        out = 1/numpy.sqrt(2*numpy.pi*x)/x*numpy.exp(-1/(2*x))
        out[x == 0] = 0
        return out

    def _cdf(self, x):
        return 2*(1-special.ndtr(1/numpy.sqrt(x)))

    def _ppf(self, q):
        val = special.ndtri(1-q/2.0)
        return 1.0/(val*val)

    def _upper(self):
        return 1e12

    def _lower(self):
        return 0.


class Levy(ShiftScaleDistribution):
    """
    Levy distribution

    Args:
        scale (float, Distribution):
            Scaling parameter
        shift (float, Distribution):
            Location parameter

    Examples:
        >>> distribution = chaospy.Levy()
        >>> distribution
        Levy()
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc[:5].round(3)
        array([ 0.   ,  0.609,  1.412,  3.636, 15.58 ])
        >>> distribution.upper
        array([1.e+12])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0.   , 0.369, 0.167, 0.05 , 0.006, 0.   ])
        >>> distribution.sample(4).round(3)
        array([  4.965,   0.403, 257.22 ,   2.025])

    """

    def __init__(self, scale=1, shift=0):
        super(Levy, self).__init__(
            dist=levy(),
            scale=scale,
            shift=shift,
        )



import numpy
import chaospy

from ..baseclass import Distribution


class OperatorDistribution(Distribution):

    def __init__(self, left, right, exclusion=None, repr_args=None):
        if not isinstance(left, Distribution):
            left = numpy.atleast_1d(left)
            if left.ndim > 1:
                raise chaospy.UnsupportedFeature(
                    "distribution operators limited to at-most 1D arrays.")
        if not isinstance(right, Distribution):
            right = numpy.atleast_1d(right)
            if right.ndim > 1:
                raise chaospy.UnsupportedFeature(
                    "distribution operators limited to at-most 1D arrays.")
        dependencies, parameters, rotation = chaospy.declare_dependencies(
            distribution=self,
            parameters=dict(left=left, right=right),
            is_operator=True,
        )

        super(OperatorDistribution, self).__init__(
            parameters=parameters,
            dependencies=dependencies,
            exclusion=exclusion,
            repr_args=repr_args,
        )
        self._cache_copy = {}
        self._lower_cache = {}
        self._upper_cache = {}

    def get_parameters(self, idx, cache, assert_numerical=True):
        parameters = super(OperatorDistribution, self).get_parameters(
            idx, cache, assert_numerical=assert_numerical)
        assert set(parameters) == {"cache", "left", "right", "idx"}

        if isinstance(parameters["left"], Distribution):
            parameters["left"] = parameters["left"]._get_cache(idx, cache=parameters["cache"], get=0)
        elif len(parameters["left"]) > 1 and idx is not None:
            parameters["left"] = parameters["left"][idx]
        if isinstance(parameters["right"], Distribution):
            parameters["right"] = parameters["right"]._get_cache(idx, cache=parameters["cache"], get=0)
        elif len(parameters["right"]) > 1 and idx is not None:
            parameters["right"] = parameters["right"][idx]

        if assert_numerical:
            assert (not isinstance(parameters["left"], Distribution) or
                    not isinstance(parameters["right"], Distribution))
        if cache is not self._cache_copy:
            self._cache_copy = cache
            self._lower_cache = {}
            self._upper_cache = {}
        if idx is None:
            del parameters["idx"]
        return parameters



"""Discretized Stieltjes' method."""
import numpy
import numpoly


def discretized_stieltjes(order, abscissas, weights, normed=False):
    """
    Discretized Stieltjes' method.

    Args:
        order (int):
            The polynomial order create.
        abscissas (numpy.ndarray):
            Quadrature abscissas, assumed to be of ``shape == (D, N)``, where
            ``D`` is the number of distributions to handle at once, and ``N``
            is the number of abscissas.
        weights (numpy.ndarray):
            Quadrature weights, assumed to be of ``shape == (N,)``.
        normed (bool):
            If true, normalize polynomials.

    Returns:
        (numpy.ndarray, numpy.ndarray, numpy.ndarray):
            coefficients:
                The recurrence coefficients created using the discretized
                Stieltjes' method, with ``shape == (2, D, order+1)``.
            polynomials:
                The orthogonal polynomial expansion created as a by-product of
                the algorithm.
            norms:
                The norm of each orthogonal polynomial. Roughly equivalent to
                ``chaospy.E(polynomials**2, dist)``, but more numerically
                stable than most alternatives.

    Examples:
        >>> dist = chaospy.J(chaospy.Uniform(0, 1), chaospy.Beta(3, 4))
        >>> abscissas, weights = chaospy.generate_quadrature(
        ...     9, dist, rule="clenshaw_curtis")
        >>> coeffs, orth, norms = discretized_stieltjes(2, abscissas, weights)
        >>> coeffs.round(5)
        array([[[0.5    , 0.5    , 0.5    ],
                [0.42857, 0.46032, 0.47525]],
        <BLANKLINE>
               [[1.     , 0.08333, 0.06667],
                [1.     , 0.03061, 0.04321]]])
        >>> orth[2].round(5)
        polynomial([q0**2-q0+0.16667, q1**2-0.88889*q1+0.16667])
        >>> norms.round(5)
        array([[1.     , 0.08333, 0.00556],
               [1.     , 0.03061, 0.00132]])
        >>> coeffs, orth, norms = discretized_stieltjes(
        ...     2, abscissas, weights, normed=True)
        >>> coeffs.round(5)
        array([[[0.5    , 0.04167, 0.26424],
                [0.42857, 0.01409, 0.31365]],
        <BLANKLINE>
               [[1.     , 1.     , 1.     ],
                [1.     , 1.     , 1.     ]]])
        >>> orth[2].round(5)
        polynomial([3.9155*q0**2-2.1209*q0-1.04874,
                    5.94906*q1**2-2.63343*q1-1.00494])
        >>> norms.round(5)
        array([[1., 1., 1.],
               [1., 1., 1.]])
    """
    abscissas = numpy.asfarray(abscissas)
    weights = numpy.asfarray(weights)
    assert len(weights.shape) == 1
    assert len(abscissas.shape) == 2
    assert abscissas.shape[-1] == len(weights)

    poly = numpoly.variable(len(abscissas))
    orth = [numpoly.repeat(0., len(abscissas)),
            numpoly.repeat(1., len(abscissas))]

    inner = numpy.sum(abscissas*weights, -1)
    norms = [numpy.ones(len(abscissas)), numpy.ones(len(abscissas))]
    coeffs = []

    for _ in range(int(order)):

        coeffs.append((inner/norms[-1], norms[-1]/norms[-2]))
        orth.append((poly-coeffs[-1][0])*orth[-1] - orth[-2]*coeffs[-1][1])

        raw_nodes = orth[-1](*abscissas)**2*weights
        inner = numpy.sum(abscissas*raw_nodes, -1)
        norms.append(numpy.sum(raw_nodes, -1))

        if normed:
            orth[-1] = orth[-1]/numpy.sqrt(norms[-1])
            norms[-1] **= 0

    coeffs.append((inner/norms[-1], norms[-1]/norms[-2]))

    coeffs = numpy.moveaxis(coeffs, 0, 2)
    norms = numpy.array(norms[1:]).T
    orth = numpoly.polynomial(orth[1:])

    return coeffs, orth, norms



def analytical_stieljes(order, dist, normed=False):
    """
    Examples:
        >>> dist = chaospy.J(chaospy.Uniform(0, 1), chaospy.Beta(3, 4))
        >>> coeffs, orth, norms = analytical_stieljes(2, dist)
        >>> coeffs.round(5)
        array([[[0.5    , 0.5    , 0.5    ],
                [0.42857, 0.46032, 0.47475]],
        <BLANKLINE>
               [[1.     , 0.08333, 0.06667],
                [1.     , 0.03061, 0.04321]]])
        >>> orth[:, 2].round(5)
        polynomial([q0**2-q0+0.16667, q1**2-0.88889*q1+0.16667])
        >>> norms.round(5)
        array([[1.     , 0.08333, 0.00556],
               [1.     , 0.03061, 0.00132]])
        >>> coeffs, orth, norms = analytical_stieljes(2, dist, normed=True)
        >>> coeffs.round(5)
        array([[[0.5    , 0.5    , 0.5    ],
                [0.42857, 0.46032, 0.47475]],
        <BLANKLINE>
               [[1.     , 0.08333, 0.06667],
                [1.     , 0.03061, 0.04321]]])
        >>> orth[:, 2].round(5)
        polynomial([13.41641*q0**2-13.41641*q0+2.23607,
                    27.49545*q1**2-24.4404*q1+4.58258])
        >>> norms.round(5)
        array([[1., 1., 1.],
               [1., 1., 1.]])
    """
    dimensions = len(dist)
    mom_order = numpy.arange(order+1).repeat(dimensions)
    mom_order = mom_order.reshape(order+1, dimensions).T
    coeffs = dist.ttr(mom_order)
    coeffs[1, :, 0] = 1.

    var = numpoly.variable(dimensions)
    orth = [numpy.zeros(dimensions), numpy.ones(dimensions)]
    for order_ in range(order):
        orth.append(
            (var-coeffs[0, :, order_])*orth[-1]-coeffs[1, :, order_]*orth[-2])
    orth = numpoly.polynomial(orth[1:]).T

    norms = numpy.cumprod(coeffs[1], 1)
    if normed:
        orth = numpoly.true_divide(orth, numpy.sqrt(norms))
        norms **= 0

    return coeffs, orth, norms



"""Distribution utility functions."""
import os
import logging
from contextlib import wraps
from itertools import permutations

import numpy
import chaospy


def check_dependencies(distribution):
    """
    Check if the dependency structure is valid.

    Rosenblatt transformations, density calculations etc. assumes that the
    input and output of transformation is the same. It also assumes that there
    is a order defined in `distribution._rotation` so an decomposition on the
    form `p(x0), p(x1|x0), ...` is possible.

    This should be checked on function calls, not init, as it does not apply to
    e.g. moment and three-terms-recurrence calculations.

    Args:
        distribution:
            The distribution to verify is correctly set up.

    Raises:
        StochasticallyDependentError:
            If invalid dependency structure is present.

    Examples:
        >>> dist = chaospy.Uniform(0, 1)
        >>> chaospy.check_dependencies(dist)
        >>> chaospy.check_dependencies(chaospy.Normal(mu=dist))
        Traceback (most recent call last):
            ...
        chaospy.StochasticallyDependentError: \
Normal(mu=Uniform(), sigma=1) has dangling dependencies
    """
    current = set()
    for idx in distribution._rotation:
        length = len(current)
        current.update(distribution._dependencies[idx])
        if len(current) != length+1:
            raise chaospy.StochasticallyDependentError(
                "%s has dangling dependencies" % distribution)


_EXCEPTION_IN_PROGRESS = False


def report_on_exception(method):
    """
    Method decorators for getting more verbose output.

    Will output the function name and call signature to logger when an
    exception is raised. But only during testing or if the environment variable
    `CHAOSPY_DEBUG=1` is set.

    Args:
        method:
            Method to be wrapped.

    Returns:
        Same as 'method', but wrapped to include exception logger.

    """
    logger = logging.getLogger(__name__)

    @wraps(method)
    def wrapper_method(self, *args, **kwargs):
        """Method to wrap 'method'."""
        global _EXCEPTION_IN_PROGRESS
        _EXCEPTION_IN_PROGRESS = False
        try:
            ret_val = method(self, *args, **kwargs)
        except Exception as err:
            if not _EXCEPTION_IN_PROGRESS:
                _EXCEPTION_IN_PROGRESS = True
                args = ",\n    ".join([repr(arg) for arg in args]+
                                    ["%s=%r" % (key, val) for key, val in kwargs.items()])
                logger.warning("failure:\n%s.%s(\n    %s\n)", self, method.__name__, args)
            else:
                logger.warning("failure: %s.%s", self, method.__name__)
            raise
        return ret_val

    if os.environ.get("CHAOSPY_DEBUG", "") == "1":
        method = wrapper_method
    return method


def shares_dependencies(*distributions):
    """
    Check if a collection of distributions shares dependencies.

    Internal dependencies are ignored here. It is only intra-dependencies that
    are checked. Anything else than a distribution is ignored.

    Args:
        distributions (Sequence[chaospy.Distribution]):
            The distributions to check shares dependencies or not.

    Returns:
        True if there are dependencies between the provided distributions.

    Examples:
        >>> dist = chaospy.Normal()
        >>> chaospy.shares_dependencies(dist, 4)
        False
        >>> chaospy.shares_dependencies(dist, chaospy.Normal())
        False
        >>> chaospy.shares_dependencies(dist, chaospy.Normal(dist))
        True

    """
    distributions = [dist for dist in distributions
                     if isinstance(dist, chaospy.Distribution)]
    if len(distributions) == 1:
        return False
    dependencies = [{dep for deps in dist._dependencies for dep in deps}
                    for dist in distributions]
    for deps1, deps2 in permutations(dependencies, 2):
        if deps1.intersection(deps2):
            return True
    return False


def declare_dependencies(
        distribution,
        parameters,
        rotation=None,
        is_operator=False,
        dependency_type="iid",
        length=None,
):
    """
    Convenience function for declaring distribution dependencies.

    Iterates through parameters to figure out what a distribution dependency
    structure should be.

    Args:
        distribution (chaospy.Distribution):
            The distributions to to declare dependencies for.
        parameters (Dict[str, Any]):
            The distribution parameters that should be included in the
            declaration.
        is_operator (bool):
            Operators do not themselves contain uncertainty, but only inherits
            from parameters and/or wrapped distribution.
        wrapper_dist (Optional[chaospy.Distribution]):
            Distributions that are thin-wrappers to some other distribution
            should inherent dependencies.
    """
    parameters = parameters.copy()
    for name, parameter in list(parameters.items()):
        if not isinstance(parameter, chaospy.Distribution):
            parameters[name] = numpy.atleast_1d(parameter)
    if length is None:
        if rotation is None:
            length = max([len(parameter) for parameter in parameters.values()]+[1])
        else:
            length = len(rotation)
    if rotation is None:
        rotation = numpy.arange(length, dtype=int)
    else:
        rotation = numpy.asarray(rotation)

    if is_operator:
        dependencies = [set() for _ in range(length)]
    else:
        dependencies = init_dependencies(distribution, rotation, dependency_type=dependency_type)

    for name, parameter in list(parameters.items()):
        if isinstance(parameter, chaospy.Distribution):
            if len(parameter) != length:
                raise chaospy.StochasticallyDependentError(
                    "dependencies must be same length as parent")
            for dep1, dep2 in zip(dependencies, parameter._dependencies):
                dep1.update(dep2)
        else:
            parameters[name] = parameter*numpy.ones(length, dtype=int)
    return dependencies, parameters, rotation


DISTRIBUTION_IDENTIFIERS = {}

def init_dependencies(
        distribution,
        rotation,
        dependency_type="iid",
):
    """
    Declare stochastic dependency to an underlying random variable.

    Args:
        distribution (chaospy.Distribution):
            Distribution to declare dependency for.
        count (int):
            The number of variables to declare.

    Returns:
        (List[Set[int]]):
            Unique integer identifiers that represents dependencies.

    """
    rotation = numpy.asarray(rotation)
    assert rotation.dtype == int and rotation.ndim == 1
    length = len(rotation)
    next_new_id = len(DISTRIBUTION_IDENTIFIERS)
    new_identifiers = numpy.arange(next_new_id, next_new_id+length, dtype=int)
    for idx in new_identifiers:
        DISTRIBUTION_IDENTIFIERS[idx] = distribution

    if dependency_type == "iid":
        dependencies = [{idx} for idx in new_identifiers[rotation]]

    elif dependency_type == "accumulate":
        accumulant = set()
        dependencies = [None]*length
        for idx in rotation:
            accumulant.add(new_identifiers[idx])
            dependencies[idx] = accumulant.copy()

    return dependencies


def format_repr_kwargs(**parameters):
    out = []

    defaults_only = True
    for name, (param, default) in list(parameters.items()):
        defaults_only &= (
            isinstance(param, (int, float)) and param == default)
        if isinstance(param, numpy.ndarray):
            parameters[name] = (param.tolist(), default)
    if defaults_only:
        return []
    return ["%s=%s" % (key, parameters[key][0]) for key in sorted(parameters)]



"""Uniform probability distribution."""
from ..baseclass import SimpleDistribution, LowerUpperDistribution


class uniform(SimpleDistribution):
    """Uniform distribution fixed on the [-1, 1] interval."""

    def __init__(self):
        super(uniform, self).__init__()

    def _pdf(self, x):
        return 0.5

    def _cdf(self, x):
        return .5*x+.5

    def _ppf(self, q):
        return 2*q-1

    def _lower(self):
        return -1.

    def _upper(self):
        return 1.

    def _mom(self, k):
        return 1./(k+1)*(k%2 == 0)

    def _ttr(self, n):
        return 0., n*n/(4.*n*n-1)


class Uniform(LowerUpperDistribution):
    r"""
    Uniform probability distribution.

    Args:
        lower (float, Distribution):
            Lower threshold of distribution. Must be smaller than ``upper``.
        upper (float, Distribution):
            Upper threshold of distribution.

    Examples:
        >>> distribution = chaospy.Uniform(2, 4)
        >>> distribution
        Uniform(lower=2, upper=4)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([2. , 2.4, 2.8, 3.2, 3.6, 4. ])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0.5, 0.5, 0.5, 0.5, 0.5, 0.5])
        >>> distribution.sample(4).round(3)
        array([3.307, 2.23 , 3.901, 2.964])
        >>> distribution.mom(1).round(4)
        3.0
        >>> distribution.ttr([0, 1, 2, 3]).round(4)
        array([[ 3.    ,  3.    ,  3.    ,  3.    ],
               [-0.    ,  0.3333,  0.2667,  0.2571]])

    """

    def __init__(self, lower=0., upper=1.):
        super(Uniform, self).__init__(
            dist=uniform(),
            lower=lower,
            upper=upper,
        )



"""
The Saltelli method

Code is built upon the code provided by Vinzenze Eck.
"""
import numpy
import numpoly
import chaospy


class Saltelli(object):
    """
    Buffer class to be able to retrieve Saltelli matrices.

    The core of the method relies on cross-combining the columns of two random
    matrices as part of a double expectation.

    Examples:
        >>> dist = chaospy.Iid(chaospy.Uniform(), 2)
        >>> generator = Saltelli(dist, 3, rule="halton")
        >>> generator[False, False].round(4)
        array([[0.875 , 0.0625, 0.5625],
               [0.5556, 0.8889, 0.037 ]])
        >>> generator[False, True].round(4)
        array([[0.875 , 0.0625, 0.5625],
               [0.4444, 0.7778, 0.2222]])
        >>> generator[True, False].round(4)
        array([[0.125 , 0.625 , 0.375 ],
               [0.5556, 0.8889, 0.037 ]])
        >>> generator[True, True].round(4)
        array([[0.125 , 0.625 , 0.375 ],
               [0.4444, 0.7778, 0.2222]])
    """

    def __init__(self, dist, samples, poly=None, rule="random"):
        """
        Initialize the matrix generator.

        dist (chaopy.Distribution):
            distribution to sample from.
        samples (int):
            The number of samples to draw for each matrix.
        poly (numpoly.ndpoly):
            If provided, evaluated samples through polynomials before returned.
        rule (str):
            Scheme for generating random samples.
        """
        self.dist = dist
        samples_ = chaospy.generate_samples(
            2*samples, domain=len(dist), rule=rule)
        self.samples1 = samples_.T[:samples].T
        self.samples2 = samples_.T[samples:].T
        self.poly = poly
        self.buffer = {}

    def get_matrix(self, indices):
        """Retrieve Saltelli matrix."""
        new = numpy.empty(self.samples1.shape)
        for idx in range(len(indices)):
            if indices[idx]:
                new[idx] = self.samples1[idx]
            else:
                new[idx] = self.samples2[idx]
        new = self.dist.inv(new)
        if isinstance(self.poly, numpoly.ndpoly) and self.poly.size:
            new = self.poly(*new)
        return new

    def __getitem__(self, indices):
        """Shortcut to `get_matrix`."""
        assert len(self.dist) == len(indices)

        # uniquify:
        key = tuple(bool(idx) for idx in indices)

        if key in self.buffer:
            matrix = self.buffer[key]
        else:
            matrix = self.get_matrix(indices)
            self.buffer[key] = matrix

        return matrix


def Sens_m_sample(poly, dist, samples, rule="random"):
    """
    First order sensitivity indices estimated using Saltelli's method.

    Args:
        poly (numpoly.ndpoly):
            If provided, evaluated samples through polynomials before returned.
        dist (chaopy.Distribution):
            distribution to sample from.
        samples (int):
            The number of samples to draw for each matrix.
        rule (str):
            Scheme for generating random samples.

    Return:
        (numpy.ndarray):
            array with `shape == (len(dist), len(poly))` where `sens[dim][pol]`
            is the first sensitivity index for distribution dimensions `dim` and
            polynomial index `pol`.

    Examples:
        >>> dist = chaospy.Iid(chaospy.Uniform(), 2)
        >>> poly = chaospy.monomial(2, 3, names=2, reverse=False)
        >>> poly
        polynomial([q0**2, q0*q1, q1**2])
        >>> Sens_m_sample(poly, dist, 10000, rule="hammersley").round(4)
        array([[0.008 , 0.0026, 0.    ],
               [0.    , 0.6464, 2.1321]])
    """
    dim = len(dist)

    generator = Saltelli(dist, samples, poly, rule=rule)

    zeros = [0]*dim
    ones = [1]*dim
    index = [0]*dim

    variance = numpy.var(generator[zeros], -1)

    matrix_0 = generator[zeros]
    matrix_1 = generator[ones]
    mean = .5*(numpy.mean(matrix_1) + numpy.mean(matrix_0))
    matrix_0 -= mean
    matrix_1 -= mean

    out = [
        numpy.mean(matrix_1*((generator[index]-mean)-matrix_0), -1) /
        numpy.where(variance, variance, 1)
        for index in numpy.eye(dim, dtype=bool)
    ]

    return numpy.array(out)


def Sens_m2_sample(poly, dist, samples, rule="random"):
    """
    Second order sensitivity indices estimated using Saltelli's method.

    Args:
        poly (numpoly.ndpoly):
            If provided, evaluated samples through polynomials before returned.
        dist (chaopy.Distribution):
            distribution to sample from.
        samples (int):
            The number of samples to draw for each matrix.
        rule (str):
            Scheme for generating random samples.

    Return:
        (numpy.ndarray):
            array with `shape == (len(dist), len(dist), len(poly))` where
            `sens[dim1][dim2][pol]` is the correlating sensitivity between
            dimension `dim1` and `dim2` and polynomial index `pol`.

    Examples:
        >>> dist = chaospy.Iid(chaospy.Uniform(), 2)
        >>> poly = chaospy.monomial(2, 3, names=2, reverse=False)
        >>> poly
        polynomial([q0**2, q0*q1, q1**2])
        >>> Sens_m2_sample(poly, dist, 10000, rule="halton").round(4)
        array([[[ 0.008 ,  0.0026,  0.    ],
                [-0.0871,  1.1516,  1.2851]],
        <BLANKLINE>
               [[-0.0871,  1.1516,  1.2851],
                [ 0.    ,  0.7981,  1.38  ]]])
    """
    dim = len(dist)

    generator = Saltelli(dist, samples, poly, rule=rule)

    zeros = [0]*dim
    ones = [1]*dim
    index = [0]*dim

    variance = numpy.var(generator[zeros], -1)

    matrix_0 = generator[zeros]
    matrix_1 = generator[ones]
    mean = .5*(numpy.mean(matrix_1) + numpy.mean(matrix_0))

    matrix_0 -= mean
    matrix_1 -= mean

    for dim1 in range(dim):

        index[dim1] = 1
        matrix = generator[index]-mean
        vals = numpy.mean(
            matrix_1*(matrix-matrix_0),
            -1,
        ) / numpy.where(variance, variance, 1)
        if not dim1:
            out = numpy.empty((dim, dim)+vals.shape)
        out[dim1, dim1] = vals

        for dim2 in range(dim1+1, dim):

            index[dim2] = 1

            matrix = generator[index]-mean

            out[dim1, dim2] = out[dim2, dim1] = numpy.mean(
                matrix_1*(matrix-matrix_0),
                -1,
            ) / numpy.where(variance, variance, 1)

            index[dim2] = 0

        index[dim1] = 0

    return out


def Sens_t_sample(poly, dist, samples, rule="random"):
    """
    Total order sensitivity indices estimated using Saltelli's method.

    Args:
        poly (numpoly.ndpoly):
            If provided, evaluated samples through polynomials before returned.
        dist (chaopy.Distribution):
            distribution to sample from.
        samples (int):
            The number of samples to draw for each matrix.
        rule (str):
            Scheme for generating random samples.

    Return:
        (numpy.ndarray):
            array with `shape == (len(dist), len(poly))` where `sens[dim][pol]`
            is the total order sensitivity index for distribution dimensions
            `dim` and polynomial index `pol`.

    Examples:
        >>> dist = chaospy.Iid(chaospy.Uniform(0, 1), 2)
        >>> poly = chaospy.monomial(2, 3, names=2, reverse=False)
        >>> poly
        polynomial([q0**2, q0*q1, q1**2])
        >>> Sens_t_sample(poly, dist, 10000, rule="halton").round(4)
        array([[ 1.    ,  0.2   , -0.3807],
               [ 0.9916,  0.9962,  1.    ]])
    """
    generator = Saltelli(dist, samples, poly, rule=rule)

    dim = len(dist)
    zeros = [0]*dim
    variance = numpy.var(generator[zeros], -1)
    return numpy.array([
        1-numpy.mean((generator[~index]-generator[zeros])**2, -1,) /
        (2*numpy.where(variance, variance, 1))
        for index in numpy.eye(dim, dtype=bool)
    ])



"""Convert three terms recurrence coefficients into quadrature rules."""
import numpy
import scipy.linalg


def coefficients_to_quadrature(coeffs):
    """
    Construct Gaussian quadrature abscissas and weights from three terms
    recurrence coefficients.

    Examples:
        >>> distribution = chaospy.Normal(0, 1)
        >>> coeffs, = chaospy.construct_recurrence_coefficients(4, distribution)
        >>> coeffs
        array([[0., 0., 0., 0., 0.],
               [0., 1., 2., 3., 4.]])
        >>> (abscissas,), (weights,) = chaospy.coefficients_to_quadrature(coeffs)
        >>> abscissas.round(4)
        array([-2.857 , -1.3556, -0.    ,  1.3556,  2.857 ])
        >>> weights.round(4)
        array([0.0113, 0.2221, 0.5333, 0.2221, 0.0113])
    """
    coeffs = numpy.asfarray(coeffs)
    if len(coeffs.shape) == 2:
        coeffs = coeffs.reshape(1, 2, -1)
    assert len(coeffs.shape) == 3, "shape %s not allowed" % coeffs.shape
    assert coeffs.shape[-1] >= 1
    abscissas = []
    weights = []
    for coeff in coeffs:

        if numpy.any(coeff[1] < 0) or numpy.any(numpy.isnan(coeff)):
            raise numpy.linalg.LinAlgError(
                "Invalid recurrence coefficients can not be used for "
                "constructing Gaussian quadrature rule")

        order = len(coeff[0])
        if order:
            bands = numpy.zeros((2, order))
            bands[0, :] = coeff[0, :order]
            bands[1, :-1] = numpy.sqrt(coeff[1, 1:order])
            vals, vecs = scipy.linalg.eig_banded(bands, lower=True)

            abscissa, weight = vals.real, vecs[0, :]**2
            indices = numpy.argsort(abscissa)
            abscissa, weight = abscissa[indices], weight[indices]

        else:
            abscissa, weight = numpy.array([coeffs[dim][0, 0]]), numpy.array([1.])

        abscissas.append(abscissa)
        weights.append(weight)

    return abscissas, weights



"""Fatigue-life distribution."""
import numpy
from scipy import special

from ..baseclass import SimpleDistribution, ShiftScaleDistribution


class fatigue_life(SimpleDistribution):
    """Fatigue-life distribution."""

    def __init__(self, c=0):
        super(fatigue_life, self).__init__(dict(c=c))

    def _pdf(self, x, c):
        output = (x+1)/(2*c*numpy.sqrt(2*numpy.pi*x**3))
        output *= numpy.exp(-(x-1)**2/(2.0*x*c**2))
        output[(x == 0) & numpy.isnan(output)] = 0
        return output

    def _cdf(self, x, c):
        out = special.ndtr(1.0/c*(numpy.sqrt(x)-1.0/numpy.sqrt(x)))
        out = numpy.where(x == 0, 0, out)
        return out

    def _ppf(self, q, c):
        tmp = c*special.ndtri(q)
        out = numpy.where(
            numpy.isfinite(tmp), 0.25*(tmp+numpy.sqrt(tmp**2+4))**2, tmp)
        return out

    def _lower(self, c):
        return (-4*c+numpy.sqrt(16*c**2+1))**2

    def _upper(self, c):
        return (4*c+numpy.sqrt(16*c**2+1))**2


class FatigueLife(ShiftScaleDistribution):
    """
    Fatigue-Life or Birmbaum-Sanders distribution

    Args:
        shape (float, Distribution):
            Shape parameter
        scale (float, Distribution):
            Scaling parameter
        shift (float, Distribution):
            Location parameter

    Examples:
        >>> distribution = chaospy.FatigueLife(0.5)
        >>> distribution
        FatigueLife(0.5)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([ 0.056,  0.659,  0.881,  1.135,  1.519, 17.944])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0.   , 0.869, 0.879, 0.682, 0.377, 0.   ])
        >>> distribution.sample(4).round(3)
        array([1.218, 0.553, 2.23 , 0.978])

    """
    def __init__(self, shape=1, scale=1, shift=0):
        super(FatigueLife, self).__init__(
            dist=fatigue_life(shape),
            scale=scale,
            shift=shift,
            repr_args=[shape],
        )
